import gc
import sys

import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn.functional as F
import torchmetrics

from scportrait.tools.ml.models import _VGG1, _VGG2, VGG1, VGG2, CAEBase, VGG2_regression


class MultilabelSupervisedModel(pl.LightningModule):
    """
    A pytorch lightning network module to use a multi-label supervised Model.

    Parameters
    ----------
    type : str, optional, default = "VGG2"
        Network architecture to used in model. Architectures are defined in scPortrait.tools.ml.models
        Valid options: "VGG1", "VGG2", "VGG1_old", "VGG2_old".
    kwargs : dict
        Additional parameters passed to the model.

    Attributes
    ----------
    network : torch.nn.Module
        The selected network architecture.
    train_metrics : torchmetrics.MetricCollection
        MetricCollection for evaluating model on training data.
    val_metrics : torchmetrics.MetricCollection
        MetricCollection for evaluating model on validation data.
    test_metrics : torchmetrics.MetricCollection
        MetricCollection for evaluating model on test data.

    Methods
    -------
    forward(x)
        perform forward pass of model.
    configure_optimizers()
        Optimization function
    on_train_epoch_end()
        Callback function after each training epoch
    on_validation_epoch_end()
        Callback function after each validation epoch
    confusion_plot(matrix)
        Generate confusion matrix plot
    training_step(batch, batch_idx)
        Perform a single training step
    validation_step(batch, batch_idx)
        Perform a single validation step
    test_step(batch, batch_idx)
        Perform a single test step
    test_epoch_end(outputs)
        Callback function after testing epochs
    """

    def __init__(self, model_type="VGG2", image_size_factor=None, **kwargs):
        super().__init__()

        self.save_hyperparameters()

        # if modeltype is availabel in hparams load from there
        if "model_type" in self.hparams:
            model_type = self.hparams["model_type"]

        # initialize metrics
        if self.hparams["num_classes"] == 2:
            task_type = "binary"
        elif self.hparams["num_classes"] > 2:
            task_type = "multiclass"
        else:
            raise ValueError("No num_classes specified in hparams")

        if image_size_factor is not None:
            if "image_size_factor" in self.hparams:
                assert (
                    self.hparams["image_size_factor"] == image_size_factor
                ), "image size factor in hparams and passed to model do not match"
            else:
                self.hparams["image_size_factor"] = image_size_factor

        # initialize metrics to track
        self.accuracy = torchmetrics.Accuracy(task=task_type, num_classes=self.hparams["num_classes"])
        self.aucroc = torchmetrics.AUROC(task=task_type, thresholds=None, num_classes=self.hparams["num_classes"])

        if model_type == "VGG1":
            self.network = VGG1(
                in_channels=self.hparams["num_in_channels"],
                cfg="B",
                dimensions=128,
                num_classes=self.hparams["num_classes"],
                image_size_factor=self.hparams["image_size_factor"],
            )

        elif model_type == "VGG2":
            self.network = VGG2(
                in_channels=self.hparams["num_in_channels"],
                cfg="B",
                dimensions=128,
                num_classes=self.hparams["num_classes"],
                image_size_factor=self.hparams["image_size_factor"],
            )

        ## add deprecated type for backward compatibility
        elif model_type == "VGG1_old":
            self.network = _VGG1(
                in_channels=self.hparams["num_in_channels"],
                cfg="B",
                dimensions=128,
                num_classes=self.hparams["num_classes"],
            )

        ## add deprecated type for backward compatibility
        elif model_type == "VGG2_old":
            self.network = _VGG2(
                in_channels=self.hparams["num_in_channels"],
                cfg="B",
                dimensions=128,
                num_classes=self.hparams["num_classes"],
            )
        else:
            sys.exit(
                "Incorrect network architecture specified. Please check that MultilabelSupervisedModel type parameter is set to key present in method."
            )

    def forward(self, x):
        return self.network(x)

    def configure_optimizers(self):
        if self.hparams["optimizer"] == "SGD":
            optimizer = torch.optim.SGD(self.parameters(), lr=self.hparams["learning_rate"])

        elif self.hparams["optimizer"] == "Adam":
            # set weight decay to 0 if not specified in hparams
            if self.hparams["weight_decay"] is None:
                self.hparams["weight_decay"] = 0
            optimizer = torch.optim.Adam(
                self.parameters(), lr=self.hparams["learning_rate"], weight_decay=self.hparams["weight_decay"]
            )

        elif self.hparams["optimizer"] == "AdamW":
            # set weight decay to 0.01 if not specified in hparams
            if self.hparams["weight_decay"] is None:
                self.hparams["weight_decay"] = 10**-2
            optimizer = torch.optim.AdamW(
                self.parameters(), lr=self.hparams["learning_rate"], weight_decay=self.hparams["weight_decay"]
            )

        else:
            raise ValueError("No optimizer specified in hparams")
        return optimizer

    def training_step(self, batch, batch_idx):
        data, label = batch

        # calculate loss
        output_softmax = self.network(data)
        loss = F.nll_loss(output_softmax, label)

        # calculate accuracy
        probabilities = torch.exp(output_softmax)
        pred_labels = torch.argmax(probabilities, dim=1)
        acc = self.accuracy(pred_labels, label)

        self.log("loss/train", loss, prog_bar=True, on_epoch=True, sync_dist=True)
        self.log("acc/train", acc, prog_bar=True, on_epoch=True, sync_dist=True)

        return {"loss": loss, "probabilities": probabilities, "actual_labels": label}

    def validation_step(self, batch, batch_idx):
        data, label = batch

        output_softmax = self.network(data)
        loss = F.nll_loss(output_softmax, label)

        # calculate accuracy
        probabilities = torch.exp(
            output_softmax
        )  # we use the logsoftmax so we need to take the exp to get the actual probabilities
        pred_labels = torch.argmax(
            probabilities, dim=1
        )  # then we can select the predicted label taking a 0.5 threshold for binary classification, for multiclass problems it simply selects the most likely class
        acc = self.accuracy(pred_labels, label)

        self.log("loss/val", loss, prog_bar=True, on_epoch=True, sync_dist=True)
        self.log("acc/val", acc, prog_bar=True, on_epoch=True, sync_dist=True)

        return {"loss": loss, "probabilities": probabilities, "actual_labels": label}

    def test_step(self, batch, batch_idx):
        data, label = batch

        output_softmax = self.network(data)
        loss = F.nll_loss(output_softmax, label)

        # calculate accuracy
        probabilities = torch.exp(output_softmax)
        pred_labels = torch.argmax(probabilities, dim=1)
        acc = self.accuracy(pred_labels, label)

        self.log("loss/test", loss, prog_bar=True, on_epoch=True, sync_dist=True)
        self.log("acc/test", acc, prog_bar=True, on_epoch=True, sync_dist=True)

        return {"loss": loss, "probabilities": probabilities, "actual_labels": label}


class RegressionModel(pl.LightningModule):
    def __init__(self, model_type="VGG2_regression", **kwargs):
        super().__init__()
        self.save_hyperparameters()

        # Define the regression model
        if model_type == "VGG2_regression":
            self.network = VGG2_regression(
                cfg="B",
                cfg_MLP="A",
                in_channels=self.hparams["num_in_channels"],
                image_size_factor=self.hparams["image_size_factor"],
            )

        # Initialize metrics for regression model
        self.mse = torchmetrics.MeanSquaredError()  # MSE metric for regression
        self.mae = torchmetrics.MeanAbsoluteError()  # MAE metric for regression

    def forward(self, x):
        return self.network(x)

    def configure_optimizers(self):
        if self.hparams["optimizer"] == "SGD":
            optimizer = torch.optim.SGD(self.parameters(), lr=self.hparams["learning_rate"])

        elif self.hparams["optimizer"] == "Adam":
            if self.hparams["weight_decay"] is None:
                self.hparams["weight_decay"] = 0
            optimizer = torch.optim.Adam(
                self.parameters(), lr=self.hparams["learning_rate"], weight_decay=self.hparams["weight_decay"]
            )

        elif self.hparams["optimizer"] == "AdamW":
            if self.hparams["weight_decay"] is None:
                self.hparams["weight_decay"] = 10**-2
            optimizer = torch.optim.AdamW(
                self.parameters(), lr=self.hparams["learning_rate"], weight_decay=self.hparams["weight_decay"]
            )

        else:
            raise ValueError("No optimizer specified in hparams.")

        return optimizer

    def configure_loss(self):
        if self.hparams["loss"] == "mse":
            loss = F.mse_loss
        elif self.hparams["loss"] == "huber":
            if self.hparams["huber_delta"] is None:
                self.hparams["huber_delta"] = 1.0
            loss = F.huber_loss
        else:
            raise ValueError("No loss function specified in hparams.")

        return loss

    def training_step(self, batch):
        data, target = batch
        target = target.unsqueeze(1)
        output = self.network(data)  # Forward pass, only one output

        loss_func = self.configure_loss()

        if self.hparams["loss"] == "huber":
            loss = loss_func(output, target, delta=self.hparams["huber_delta"], reduction="mean")
        else:
            loss = loss_func(output, target)

        self.log("loss/train", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("mse/train", self.mse(output, target), on_epoch=True, prog_bar=True)
        self.log("mae/train", self.mae(output, target), on_epoch=True, prog_bar=True)

        return {"loss": loss, "predictions": output, "targets": target}

    def validation_step(self, batch):
        data, target = batch
        target = target.unsqueeze(1)
        output = self.network(data)

        loss_func = self.configure_loss()

        if self.hparams["loss"] == "huber":
            loss = loss_func(output, target, delta=self.hparams["huber_delta"], reduction="mean")
        else:
            loss = loss_func(output, target)

        self.log("loss/val", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("mse/val", self.mse(output, target), on_epoch=True, prog_bar=True)
        self.log("mae/val", self.mae(output, target), on_epoch=True, prog_bar=True)

        return {"loss": loss, "predictions": output, "targets": target}

    def test_step(self, batch):
        data, target = batch
        target = target.unsqueeze(1)
        output = self.network(data)

        loss_func = self.configure_loss()

        if self.hparams["loss"] == "huber":
            loss = loss_func(output, target, delta=self.hparams["huber_delta"], reduction="mean")
        else:
            loss = loss_func(output, target)

        self.log("loss/test", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("mse/test", self.mse(output, target), on_epoch=True, prog_bar=True)
        self.log("mae/test", self.mae(output, target), on_epoch=True, prog_bar=True)

        return {"loss": loss, "predictions": output, "targets": target}


# implemented models for future use currently not applied to scPortrait


class GeneralModel(pl.LightningModule):
    def __init__(self, model, hparams):
        super().__init__()

        print(hparams)
        self.hp = hparams
        self.network = model

        self.train_metrics = torchmetrics.MetricCollection(
            [
                torchmetrics.Precision("binary", average="none", num_classes=hparams["num_classes"]),
                torchmetrics.Recall("binary", average="none", num_classes=hparams["num_classes"]),
                torchmetrics.AUROC(num_classes=hparams["num_classes"]),
                torchmetrics.Accuracy("binary"),
            ]
        )

        self.val_metrics = torchmetrics.MetricCollection(
            [
                torchmetrics.Precision("binary", average="none", num_classes=hparams["num_classes"]),
                torchmetrics.Recall("binary", average="none", num_classes=hparams["num_classes"]),
                torchmetrics.AUROC(num_classes=hparams["num_classes"]),
                torchmetrics.Accuracy("binary"),
            ]
        )

    def on_train_start(self):
        self.logger.log_hyperparams(
            self.hp, {"precision/train": 0, "recall/train": 0, "precision/val": 0, "recall/val": 0}
        )

    def forward(self, x):
        return self.network(x)

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.parameters(), lr=self.hp["lr"])
        return optimizer

    def on_train_epoch_start(self):
        pass

    def on_validation_epoch_start(self):
        pass

    def on_train_epoch_end(self, outputs):
        metrics = self.train_metrics.compute()

        self.log("precision/train", metrics["Precision"][0])
        self.log("recall/train", metrics["Recall"][0])
        self.log("auroc/train", metrics["AUROC"])
        self.log("accuracy/train", metrics["Accuracy"])

        # Reseting internal state such that metric ready for new data
        self.train_metrics.reset()

    def on_validation_epoch_end(self):
        metrics = self.val_metrics.compute()

        self.log("precision/val", metrics["Precision"][0])
        self.log("recall/val", metrics["Recall"][0])
        self.log("auroc/val", metrics["AUROC"])
        self.log("accuracy/val", metrics["Accuracy"])

        # Reseting internal state such that metric ready for new data
        self.val_metrics.reset()

    def training_step(self, batch, batch_idx):
        data, label = batch

        data, label = data.cuda(), label.cuda()

        output = self.network(data)
        loss = F.nll_loss(output, label)

        # log accuracy metrics
        non_log = torch.exp(output)

        self.train_metrics(non_log, label)

        self.log("loss/train", loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        data, label = batch
        data, label = data.cuda(), label.cuda()

        output = self.network(data)
        loss = F.nll_loss(output, label)

        # accuracy metrics
        non_log = torch.exp(output)

        self.val_metrics(non_log, label)

        self.log("loss/val", loss, prog_bar=True)


class AutoEncoderModel(pl.LightningModule):
    def __init__(self, **kwargs):
        super().__init__()

        print(self.hparams)
        self.save_hyperparameters()

        self.network = CAEBase(
            encoder_cfg=self.hparams["encoder_cfg"],
            decoder_cfg=self.hparams["decoder_cfg"],
            in_channels=self.hparams["num_in_channels"],
            out_channels=self.hparams["num_out_channels"],
        )

        self.loss = torch.nn.BCELoss()

        # Important: This property activates manual optimization.
        self.automatic_optimization = False

    def forward(self, x):
        return self.network(x)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams["learning_rate"])

        return optimizer

    def on_train_epoch_start(self):
        pass

    def on_validation_epoch_start(self):
        pass

    def on_train_epoch_end(self, outputs):
        pass
        # tensorboard = self.logger.experiment
        # tensorboard.add_image()

    def training_step(self, batch, batch_idx):
        opt = self.optimizers()
        opt.zero_grad()

        images, _ = batch

        images = images.cuda()
        input_golgi_channel = images[:, 3:4, :, :]

        output = self.network(images)

        loss = self.loss(output, input_golgi_channel)

        self.manual_backward(loss, retain_graph=False)

        opt.step()
        loss = loss.detach()

        self.log("loss/train", loss, prog_bar=True)

        if batch_idx > 118:
            torch.cuda.empty_cache()

            for obj in gc.get_objects():
                try:
                    if torch.is_tensor(obj) or (hasattr(obj, "data") and torch.is_tensor(obj.data)):
                        print(type(obj), obj.size())
                except (AttributeError, TypeError):
                    pass

    def validation_step(self, batch, batch_idx):
        opt = self.optimizers()
        opt.zero_grad()

        images, labels = batch

        images = images.cuda()
        input_golgi_channel = images[:, 3:4, :, :]

        output = self.network(images)

        loss = self.loss(output, input_golgi_channel)

        loss = loss.detach()

        if batch_idx == 0:
            SAMPLE_NUM = 10

            input_sample = images[:SAMPLE_NUM, :, :, :].detach().cpu()
            output_sample = output[:SAMPLE_NUM, :, :, :].detach().cpu()
            label_sample = labels[:SAMPLE_NUM].detach().cpu()

            img = self.sample_plot(input_sample, output_sample, label_sample)
            self.logger.experiment.add_image("prediction", img, self.current_epoch, dataformats="NHWC")

        print(batch_idx, self.current_epoch)

        self.log("loss/val", loss.item(), prog_bar=True)
        return loss

    def sample_plot(self, input_sample, output_sample, label_sample):
        fig, axs = plt.subplots(6, len(label_sample), figsize=(15, 10))

        for i in range(len(label_sample)):
            label = int(label_sample[i].item())

            channels = input_sample[i]

            axs[0, i].set_title(self.hparams["class_labels"][label])

            for j in range(6):
                if j < 2:
                    color = "gray"
                else:
                    color = "magma"

                if j < 5:
                    c = channels[j]

                else:
                    c = output_sample[i, 0]

                axs[j, i].imshow(c, cmap=color)
                axs[j, i].xaxis.set_ticks([])
                axs[j, i].yaxis.set_ticks([])

        rows = ["Cell Mask", "Nucleus Mask", "Nucleus", "TGOLN2-mCherry", "WGA", "CAE"]

        for ax, row in zip(axs[:, 0], rows, strict=False):
            ax.set_ylabel(row, rotation=0, size="large", ha="right")

        plt.subplots_adjust(wspace=0.1, hspace=0.1)

        fig.tight_layout()

        fig.canvas.draw()

        data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        data = np.expand_dims(data, axis=0)

        return data


class AEModel(pl.LightningModule):
    def __init__(self, model, hparams):
        super().__init__()

        print(hparams)
        self.hp = hparams
        self.network = model
        self.loss = torch.nn.BCELoss()

        # Important: This property activates manual optimization.
        self.automatic_optimization = False

    def forward(self, x):
        return self.network(x)

    def configure_optimizers(self):
        optimizer = optimizer = torch.optim.Adam(self.parameters(), lr=self.hp["lr"])

        return optimizer

    def on_train_epoch_start(self):
        pass

    def on_validation_epoch_start(self):
        pass

    def on_train_epoch_end(self, outputs):
        pass
        # tensorboard = self.logger.experiment
        # tensorboard.add_image()

    def training_step(self, batch, batch_idx):
        opt = self.optimizers()
        opt.zero_grad()

        images, _ = batch

        images = images.cuda()
        input_golgi_channel = images[:, 3:4, :, :]

        output = self.network(images)

        loss = self.loss(output, input_golgi_channel)
        self.manual_backward(loss)

        opt.step()
        loss = loss.detach()

        self.log("loss/train", loss, prog_bar=True)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import h5py
import matplotlib.pyplot as plt
import numpy as np
from skimage.color import label2rgb

if TYPE_CHECKING:
    from scportrait.pipeline.project import Project


def plot_image(
    array: np.ndarray,
    size: tuple[int, int] = (10, 10),
    save_name: str | None = "",
    cmap: str = "magma",
    return_fig: bool = False,
    **kwargs,
):
    """
    Visualize and optionally save an input array as an image.

    This function displays a 2D array using matplotlib and can save
    the resulting image as a PNG file.

    Args:
        array: Input 2D numpy array to be plotted.
        size: Figure size in inches, by default (10, 10).
        save_name: Name of the output file, without extension. If not provided, image will not be saved, by default "".
        cmap: Color map used to display the array, by default "magma".
        **kwargs: Additional keyword arguments to be passed to `ax.imshow`.

    Returns:
        None: The function will display the image but does not return any values.

    Example:
    >>> array = np.random.rand(10, 10)
    >>> plot_image(array, size=(5, 5))
    """

    fig = plt.figure(frameon=False)
    fig.set_size_inches(size)
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.imshow(array, cmap=cmap, **kwargs)

    if return_fig:
        return fig

    if save_name != "":
        plt.savefig(save_name + ".png")
        plt.show()
        plt.close()
    else:
        plt.show()
        plt.close()


def visualize_class(
    class_ids: np.ndarray | list[int], seg_map: np.ndarray, image: np.ndarray, all_ids=None, return_fig=False, **kwargs
):
    """
    Visualize specific classes in a segmentation map by highlighting them on top of a background image.

    This function takes in class IDs and a segmentation map, and creates an output visualization
    where the specified classes are highlighted on top of the provided background image.

    Args:
        class_ids: A list or array of integers representing the class IDs to be highlighted.
        seg_map: A 2D array representing the segmentation map, where each value corresponds to a class ID.
        background: Background image (2D or 3D) on which the classes will be highlighted. Its size should match that of `seg_map`.
        *args: Any additional positional arguments that are passed to the underlying plotting functions.
        **kwargs: Any additional keyword arguments that are passed underlying plotting functions.

    Returns:
        None: The function will display the highlighted image but does not return any values.

    Example:
    >>> class_ids = [1, 2]
    >>> seg_map = np.array([[0, 1, 0], [1, 2, 1], [2, 0, 1]])
    >>> background = np.random.random((3, 3)) * 255
    >>> visualize_class(class_ids, seg_map, background)
    """
    # ensure the class ids are a list
    if not isinstance(class_ids, list):
        class_ids = list(class_ids)

    if all_ids is None:
        all_ids = set(np.unique(seg_map)) - {0}

    # get the ids to keep
    keep_ids = list(all_ids - set(class_ids))

    mask_discard = np.isin(seg_map, keep_ids)
    mask_keep = np.isin(seg_map, class_ids)

    # Set the values in the output map to 2 for the specified class IDs
    outmap = np.where(mask_discard, 2, seg_map)

    # Set the values in the output map to 1 for all class IDs other than the specified classes
    outmap = np.where(mask_keep, 1, outmap)

    vis_map = label2rgb(outmap, image=image, colors=["red", "blue"], alpha=0.4, bg_label=0)

    fig = plot_image(vis_map, return_fig=True, **kwargs)

    if return_fig:
        return fig


def plot_segmentation_mask(
    project: Project,
    mask_channel: int = 0,
    image_channel: int = 0,
    selection: tuple[slice, slice] | None = None,
    cmap_image: str = "Greys_r",
    cmap_masks: str = "prism",
    alpha: float = 0.5,
    figsize: tuple[int, int] = (10, 10),
) -> object:
    """Visualize the segmentation mask overlayed with a channel of the input image.

    Args:
       project: Instance of a scPortrait project.
       mask_channel: The index of the channel to use for the segmentation mask.
       image_channel: The index of the channel to use for the image.
       selection: The selection coordinates for a specific region of interest.
       cmap_image: The colormap to use for the input image.
       cmap_masks: The colormap to use for the segmentation mask.
       alpha: The transparency level of the segmentation mask.
       figsize: The figure size as (width, height).

    Returns:
        The generated figure object.
    """
    # integer value indicating background (default value)
    background = 0

    segmentation_file = os.path.join(project.seg_directory, project.segmentation_f.DEFAULT_SEGMENTATION_FILE)

    with h5py.File(segmentation_file, "r") as hf:
        segmentation = hf.get(project.segmentation_f.DEFAULT_MASK_NAME)
        channels = hf.get(project.segmentation_f.DEFAULT_CHANNELS_NAME)

        if selection is None:
            segmentation = segmentation[mask_channel, :, :]
            image = channels[image_channel, :, :]
        else:
            segmentation = segmentation[mask_channel, selection[0], selection[1]]
            image = channels[image_channel, selection[0], selection[1]]

    # set background to np.nan so that its not visualized
    segmentation = np.where(segmentation == background, np.nan, segmentation)

    fig = plt.figure(figsize=figsize)
    plt.imshow(image, cmap=cmap_image)
    plt.imshow(segmentation, alpha=alpha, cmap=cmap_masks)
    plt.axis("off")
    return fig


def colorize(
    im: np.ndarray, color: tuple[int, ...] = (1, 0, 0), clip_percentile: float = 0.0, normalize_image: bool = False
):
    """
    Helper function to create an RGB image from a single-channel image using a
    specific color.

    Args:
        im: A single-channel input image. If normalize_image = False, ensure that its values fall between the [0, 1] range.
        color: The color to use for the image. Defaults to (1, 0, 0).
        clip_percentile: The percentile to clip the image at rescaling. Defaults to 0.0 which is equivalent to Min-Max scaling.
        normalize_image: boolean value indicating if rescaling should be performed or not.

    Returns:
        np.ndarray: The colorized image.
    """
    # Check that we do just have a 2D image
    if im.ndim > 2 and im.shape[2] != 1:
        raise ValueError("This function expects a single-channel image!")

    # Rescale the image according to how we want to display it
    if normalize_image:
        im_scaled = im.astype(np.float32) - np.percentile(im, clip_percentile)
        im_scaled = im_scaled / np.percentile(im_scaled, 100 - clip_percentile)
        im_scaled = np.clip(im_scaled, 0, 1)
    else:
        im_scaled = im.astype(np.float32)
        # ensure that its 0, 1 limited otherwise it will generate errors

        assert im_scaled.min() >= 0.0, "If normalize_image = False, expected 0, 1 ranged input images"
        assert im_scaled.max() <= 1.0, "If normalize_image = False, expected 0, 1 ranged input images"

    # Need to make sure we have a channels dimension for the multiplication to work
    im_scaled = np.atleast_3d(im_scaled)

    # Reshape the color (here, we assume channels last)
    color = np.asarray(color).reshape((1, 1, -1))
    return im_scaled * color


def generate_composite(images: np.ndarray, colors: list[tuple[int, ...]] = None, plot: bool = False) -> np.ndarray:
    """Create a composite image from a multi-channel image for visualization.

    Args:
        images: A multi-channel image to be combined.
        colors: A list of colors to use for each channel. Defaults to None.
        plot: Whether to plot the composite image. Defaults to False.

    Returns:
        The composite image.
    """
    if colors is None:
        colors = [(0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 0, 1)]
    colorized = []
    for image, color in zip(images, colors, strict=False):
        image = colorize(image, color, 0.0)
        colorized.append(image)

    if plot:
        for i in colorized:
            plt.figure()
            plt.imshow(i)

    image = colorized[0]
    for i in range(len(colorized) - 1):
        image += colorized[i + 1]

    return np.clip(image, 0, 1)



from pathlib import PosixPath
from typing import TypeVar

import yaml

T = TypeVar("T")


def read_config(config_path: str | PosixPath) -> dict:
    with open(config_path) as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    return config


class QuotedStringDumper(yaml.SafeDumper):
    def represent_str(self, data):
        return self.represent_scalar("tag:yaml.org,2002:str", data, style='"')


QuotedStringDumper.add_representer(str, QuotedStringDumper.represent_str)


def write_config(config: dict, config_path: str | PosixPath) -> None:
    with open(config_path, "w") as stream:
        try:
            yaml.dump(config, stream, sort_keys=False, Dumper=QuotedStringDumper)
        except yaml.YAMLError as exc:
            print(exc)


def flatten(nested_list: list[list[T]]) -> list[T | tuple[T]]:
    """Flatten a list of lists into a single list.

    Args:
        nested_list: A list containing one or more lists as its elements

    Returns:
        A single list containing all elements from the input lists

    Example:
        >>> nested_list = [[1, 2, 3], [4, 5], [6, 7, 8, 9]]
        >>> flatten(nested_list)
        [1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return [item for sublist in nested_list for item in sublist]


def _check_for_spatialdata_plot() -> None:
    """Helper function to check if required package is installed"""
    # check for spatialdata_plot
    try:
        import spatialdata_plot
    except ImportError:
        raise ImportError(
            "Extended plotting capabilities required. Please install with `pip install 'scportrait[plotting]'`."
        ) from None



from pathlib import Path
from typing import Literal

from scportrait.data._dataloader import _download
from scportrait.tools.ml.pretrained_models import get_data_dir


def _get_remote_dataset(
    dataset: str,
    url: str,
    name: str | None = None,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] | None = "zip",
    outfile_name: None | str = None,
    outdirectory: None | str = None,
    force_download: bool = False,
) -> Path:
    """Download and extract a dataset from a remote location.

    Args:
        dataset: Name of the folder to save the dataset
        url: URL to the dataset
        name: Name of a specific file in the dataset folder to return, if None the whole dataset folder is returned

    Returns:
        Path to the downloaded and extracted dataset or a specific file in the dataset
    """
    data_dir = get_data_dir()
    if outdirectory is not None:
        data_dir = data_dir / outdirectory
        data_dir.mkdir(parents=True, exist_ok=True)
    save_path = data_dir / dataset

    if force_download:
        _download(url=url, output_path=str(save_path), output_file_name=outfile_name, archive_format=archive_format)
    elif not save_path.exists():
        _download(url=url, output_path=str(save_path), output_file_name=outfile_name, archive_format=archive_format)
    if name is None:
        return save_path
    else:
        return save_path / name


def get_config_file(config_id, force_download: bool = False) -> Path:
    config_files = {
        "dataset_1_wga_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_1/config_example1_WGASegmentation.yml",
        "dataset_1_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_1/config_example1.yml",
        "dataset_1_custom_cellpost_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_1/config_example1_custom_params.yml",
        "dataset_2_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_2/config_example2.yml",
        "dataset_3_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_3/config_example3.yml",
        "dataset_4_config": "https://raw.githubusercontent.com/MannLabs/scPortrait-notebooks/main/example_projects/example_4/config_example4.yml",
    }

    DATASET = config_id
    URL = config_files[config_id]
    NAME = "config.yml"
    return _get_remote_dataset(
        DATASET,
        URL,
        NAME,
        archive_format=None,
        outfile_name=NAME,
        outdirectory="example_configs",
        force_download=force_download,
    )


def custom_cellpose_model() -> Path:
    """Download the example custom cellpose model.

    Returns:
        Path to the downloaded and extracted custom cellpose model
    """
    DATASET = "custom_cellpose_model"
    URL = "https://zenodo.org/records/14931602/files/custom_cellpose_model.cpkt?download=1"
    NAME = "custom_cellpose_model.cpkt"
    return _get_remote_dataset(DATASET, URL, NAME, archive_format=None, outfile_name=NAME)


def autophagosome_h5sc() -> list[Path]:
    """Download the example autophagosome h5sc datasets.

    Consists of two h5sc files with a small subset of single-cell images of cells with and without autophagosomes.
    The first file is autophagy positive. The second file is autophagy negative.

    Returns:
        Path to the downloaded and extracted h5sc datasets
    """
    DATASET = "autophagosome_h5sc"
    URL = "https://zenodo.org/api/records/15105848/files-archive"
    NAMES = ["stimulated_small.h5sc", "unstimulated_small.h5sc"]
    save_path = _get_remote_dataset(DATASET, URL)
    return [save_path / name for name in NAMES]


def dataset_1() -> Path:
    """Download and extract the example dataset 1 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_1_images"
    URL = "https://zenodo.org/records/13385933/files/example_1_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_1_omezarr() -> Path:
    """Download and extract the example dataset 1 images in ome.zarr format.

    Returns:
        Path to the downloaded ome.zarr file.
    """
    DATASET = "example_1_images_omezarr"
    URL = "https://zenodo.org/records/14841309/files/input_image.ome.zarr.zip?download=1"
    NAME = "input_image.ome.zarr"
    return _get_remote_dataset(DATASET, URL, NAME)


def dataset_2() -> Path:
    """Download and extract the example dataset 2 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_2_images"
    URL = "https://zenodo.org/records/13742316/files/example_2_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_3() -> Path:
    """Download and extract the example dataset 3 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_3_images"
    URL = "https://zenodo.org/records/13742319/files/example_3_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_4() -> Path:
    """Download and extract the example dataset 4 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_4_images"
    URL = "https://zenodo.org/records/13742331/files/example_4_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_5() -> Path:
    """Download and extract the example dataset 5 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_5_images"
    URL = "https://zenodo.org/records/13742344/files/example_5_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_6() -> Path:
    """Download and extract the example dataset 6 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_6_images"
    URL = "https://zenodo.org/records/13742373/files/example_6_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_7() -> Path:
    """Download and extract the example dataset 7 images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "example_7_images"
    URL = "https://zenodo.org/records/14951047/files/example_7_images.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_stitching_example() -> Path:
    """Download and extract the example dataset for stitching images.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "stitching_example"
    URL = "https://zenodo.org/records/13742379/files/example_stitching.zip?download=1"
    return _get_remote_dataset(DATASET, URL)


def dataset_parsing_example() -> Path:
    """Download and extract the example dataset for parsing Harmony exported imaging experiments.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "parsing_example_basic"
    URL = "https://zenodo.org/records/14193689/files/harmony_export_V7_basic.zip?download=1"
    NAME = "basic_export"
    return _get_remote_dataset(DATASET, URL, NAME)


def dataset_parsing_example_flatfield_corrected() -> Path:
    """Download and extract the example dataset for parsing Harmony exported imaging experiments.

    Returns:
        Path to the downloaded and extracted images.
    """
    DATASET = "parsing_example_flatfield"
    URL = "https://zenodo.org/records/14973295/files/harmony_export_V7_flatfield_corrected.zip?download=1"
    NAME = "flat_field_corrected"
    return _get_remote_dataset(DATASET, URL, NAME)



try:
    from graph_tool import Graph
    from graph_tool.topology import shortest_distance
except ImportError:
    Graph = None
    shortest_distance = None


# calcultion of centers is equivalent to method in networkx: https://networkx.org/documentation/stable/reference/algorithms/generated/networkx.algorithms.distance_measures.center.html
def get_center_nodes(g):
    """Calculate the center node of the graph.

    The center is the set of nodes with eccentricity equal to radius.

    Parameters
    ----------
    g : graph-tool Graph
        The graph for which to calculate the center node.

    Returns
    -------
    centers : int
        index of the center node of the graph.
    """
    if shortest_distance is None:
        raise ImportError("graph-tool is required for this function")

    nodes = g.get_vertices()

    if len(nodes) > 2:
        distances = shortest_distance(g)
        eccentricities = [x.a.max() for x in distances]

        centers = nodes[eccentricities.index(min(eccentricities))]

        return centers

    # if only one node or two nodes are present in the graph the one with the lower index is returned
    else:
        return nodes[0]


def nx2gt(nxG):
    """
    Converts a networkx graph to a graph-tool graph.

    Parameters
    ----------
    nxG : networkx.Graph
        The networkx graph to convert.

    Returns
    -------
    gtG : graph-tool.Graph
        The graph-tool graph.
    """
    node_list = list(nxG.nodes())
    edge_list = list(nxG.edges(data=True))
    edge_list = [(x, y, weight["weight"]) for x, y, weight in edge_list]
    gtG = Graph(edge_list, eprops=[("weight", "float")], directed=nxG.is_directed())

    vertices = gtG.get_vertices()
    for _missing_node in [x for x in node_list if x not in vertices]:
        gtG.add_vertex()

    return gtG


def gt2nx(gtG):
    """
    Converts a graph-tool graph to a NetworkX graph.

    Parameters
    ----------
    gtG : graph-tool.Graph
        The graph-tool graph to convert.

    Returns
    -------
    nxG : networkx.Graph
        The NetworkX graph.
    """
    try:
        from networkx import Graph as nxGraph
    except ImportError:
        raise ImportError(
            "networkx is required to use stitching capabilities. Please install with 'pip install networkx'"
        ) from None

    # Create a new NetworkX graph
    nxG = nxGraph()
    edges = gtG.get_edges()
    nxG.add_edges_from(edges)
    return nxG



import os
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import norm
from skimage.morphology import dilation, disk, erosion
from sklearn.mixture import GaussianMixture

from scportrait.pipeline._base import Logable

# for visualization purposes
from scportrait.plotting._utils import _custom_cmap
from scportrait.processing.images._image_processing import downsample_img_pxs


class _BaseFilter(Logable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_unique_ids(self, mask):
        return np.unique(mask)[1:]  # to remove the background

    def get_updated_mask(self, mask: np.ndarray, ids_to_remove: np.ndarray) -> np.ndarray:
        """Update the given mask by setting the values corresponding to the specified IDs to 0.

        Args:
            mask: The mask to be updated.
            ids_to_remove: The IDs to be removed from the mask.

        Returns:
            The updated mask with the specified IDs set to 0.
        """
        return np.where(np.isin(mask, ids_to_remove), 0, mask)

    def get_downsampled_mask(self, mask):
        return downsample_img_pxs(mask, N=self.downsampling_factor)

    def get_upscaled_mask_basic(self, mask, erosion_dilation=False):
        mask = mask.repeat(self.downsampling_factor, axis=0).repeat(self.downsampling_factor, axis=1)

        if erosion_dilation:
            mask = erosion(mask, footprint=disk(self.smoothing_kernel_size))
            mask = dilation(mask, footprint=disk(self.smoothing_kernel_size))

        return mask


class SizeFilter(_BaseFilter):
    """Filter class for removing objects from a mask based on their size.

    This class provides methods to remove objects from a segmentation mask based on their size.
    If specified the objects are filtered using a threshold range passed by the user. Otherwise,
    this threshold range will be automatically calculated.

    To automatically calculate the threshold range, a gaussian mixture model will be fitted to the data.
    Per default, the number of components is set to 2, as it is assumed that the objects in the mask can be divided into two
    groups: small and large objects. The small objects constitute segmentation artefacts (partial masks that are
    frequently generated by segmentation models like e.g. cellpose) while the large objects
    represent the actual cell masks of interest. Using the fitted model, the filtering thresholds are calculated
    to remove all cells that fall outside of the given confidence interval.

    Args:
        filter_threshold: The lower and upper thresholds for object size filtering. If not provided, it will be automatically calculated.
        label: The label of the mask.
        log: Whether to take the logarithm of the size of the objects before fitting the normal distribution.
            By enabling this option, the filter will better be able to distinguish between small and large objects.
        plot_qc: Whether to plot quality control figures.
        directory: The directory to save the generated figures. If not provided, the current working directory will be used.
        confidence_interval: The confidence interval for calculating the filtering threshold.
        n_components: The number of components in the Gaussian mixture model.
        population_to_keep: For multipopulation models this parameter determines which population should be kept.
            Options are "largest", "smallest", "mostcommon", "leastcommon".
            If set to "largest" or "smallest", the model is chosen which has the largest or smallest mean.
            If set to "mostcommon" or "leastcommon", the model is chosen whose population is least or most common.
        filter_lower: Whether to filter objects that are smaller than the lower threshold.
        filter_upper: Whether to filter objects that are larger than the upper threshold.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.

     Examples
     >>> # Create a SizeFilter object
     >>> filter = SizeFilter(filter_threshold=(100, 200), label="my_mask")
     >>> # Apply the filter to a mask
     >>> filtered_mask = filter.filter(input_mask)
     >>> # Get the object IDs to be removed
     >>> ids_to_remove = filter.get_ids_to_remove(input_mask)
     >>> # Update the mask by removing the identified object IDs
     >>> updated_mask = filter.update_mask(input_mask, ids_to_remove)
    """

    def __init__(
        self,
        filter_threshold: tuple[float, float] | None = None,
        label: str = "segmask",
        log: bool = True,
        plot_qc: bool = True,
        directory: str | None = None,
        confidence_interval: float = 0.95,
        n_components: int = 1,
        population_to_keep: str = "largest",
        filter_lower: bool = True,
        filter_upper: bool = True,
        downsampling_factor: int | None = None,
        erosion_dilation: bool = True,
        smoothing_kernel_size: int = 7,
        *args,
        **kwargs,
    ):
        if directory is not None:
            self.directory = directory
        else:
            self.directory = os.getcwd()

        super().__init__(*args, directory=self.directory, **kwargs)

        self.log_values = log
        self.plot_qc = plot_qc
        self.label = label
        self.filter_threshold = filter_threshold
        self.confidence_interval = confidence_interval
        self.n_components = n_components
        self.population_to_keep = population_to_keep
        self.filter_lower = filter_lower
        self.filter_upper = filter_upper

        if downsampling_factor is not None:
            self.downsample = True
            self.downsampling_factor = downsampling_factor
            self.erosion_dilation = erosion_dilation
            self.smoothing_kernel_size = smoothing_kernel_size
        else:
            self.downsample = False

        # sanity check to ensure that some filtering will be performed
        if not self.filter_lower and not self.filter_upper:
            raise ValueError(
                "At least one of filter_lower or filter_upper must be True otherwise no filtering will be performed."
            )

        # if no directory is provided, use the current working directory

        # initialize empty placeholders for results
        self.ids_to_remove = None
        self.mask = None

    def load_mask(self, mask: np.ndarray) -> None:
        """Load the mask to be filtered.

        Args:
            mask: The mask to be filtered.
        """
        if self.downsample:
            if self.mask is None:
                self.mask = self.get_downsampled_mask(mask)
        else:
            if self.mask is None:
                self.mask = mask

    def _get_gaussian_model_plot(
        self,
        counts: np.ndarray,
        means: np.ndarray,
        variances: np.ndarray,
        weights: np.ndarray,
        threshold: tuple[float, float],
        bins: int = 30,
        figsize: tuple[int, int] = (5, 5),
        alpha: float = 0.5,
        save_figure: bool = True,
    ) -> plt.Figure:
        """Plot a histogram of the provided data with fitted Gaussian distributions.

        Args:
            counts: The input data array.
            means: The means of the Gaussian distributions.
            variances: The variances of the Gaussian distributions.
            weights: The weights of the Gaussian distributions.
            threshold: The lower and upper threshold values.
            bins: The number of bins in the histogram.
            figsize: The size of the figure.
            alpha: The transparency of the histogram bars.
            save_figure: Whether to save the figure.

        Returns:
            The generated figure.
        """

        # generate valuerange over which to visualize the distributions
        x = np.linspace(min(counts), max(counts), 1000)

        # initialize the figure
        fig, axs = plt.subplots(1, 1, figsize=figsize)

        # visualize the base histogram
        axs.hist(counts, bins=bins, density=True, alpha=alpha, label=self.label)

        # visualize the fitted Gaussian distributions
        for i in range(self.n_components):
            _pdf = norm.pdf(x, means[i], np.sqrt(variances[i])) * weights[i]
            axs.plot(x, _pdf, "-r", label=f"Gaussian {i}")

        # visualize the threshold values as dotted lines
        if self.filter_lower:
            axs.axvline(
                threshold[0],
                color="blue",
                linestyle="--",
                label=f"Threshold Lower: {threshold[0]:.2f}",
            )

        if self.filter_upper:
            axs.axvline(
                threshold[1],
                color="blue",
                linestyle="--",
                label=f"Threshold Upper: {threshold[1]:.2f}",
            )

        # format the plot
        axs.legend()
        axs.set_title("Histogram and Fitted Distributions")
        fig.tight_layout()
        plt.close(fig)

        # automatically save figure to directory if parameter is specified
        if save_figure:
            fig.savefig(os.path.join(self.directory, f"{self.label}_bimodal_model.png"))

        return fig

    def _get_histogram_plot(
        self,
        values: np.ndarray,
        label: str | None = None,
        bins: int = 30,
        alpha: float = 0.5,
        figsize: tuple[int, int] = (5, 5),
        save_figure: bool = True,
    ) -> plt.Figure:
        """Plot a histogram of the given values.

        Args:
            values: The values to be plotted.
            label: The label for the histogram plot.
            bins: The number of bins in the histogram.
            alpha: The transparency of the histogram bars.
            figsize: The size of the figure.
            save_figure: Whether to save the figure as an image.

        Returns:
            The generated figure object.
        """

        fig, axs = plt.subplots(1, 1, figsize=figsize)
        axs.hist(values, bins=bins, density=True, alpha=alpha)
        axs.set_title(label)
        axs.set_xlabel("value")
        axs.set_ylabel("density")
        fig.tight_layout()
        plt.close(fig)

        if save_figure:
            fig.savefig(os.path.join(self.directory, f"{label}_histogram.png"))

        return fig

    def _get_index_population(self, means, weights):
        """
        Returns the index of the model that matches the population of cells to be kept.

        Parameters
        ----------
        means : numpy.ndarray
            An array containing the means of the models.
        weights : numpy.ndarray
            An array containing the weights of the models.

        Returns
        -------
        int
            The index of the model that matches the population criteria.

        Notes
        -----
        The function determines the index of the model based on the population of cells that should be kept.
        The population criteria can be set to "largest", "smallest", "mostcommon", or "leastcommon".
        If the population criteria is set to "largest" or "smallest", the index is determined based on the means array.
        If the population criteria is set to "mostcommon" or "leastcommon", the index is determined based on the weights array.
        """
        if self.population_to_keep == "largest":
            idx = np.argmax(means)
        elif self.population_to_keep == "smallest":
            idx = np.argmin(means)
        elif self.population_to_keep == "mostcommon":
            idx = np.argmax(weights)
        elif self.population_to_keep == "leastcommon":
            idx = np.argmin(weights)
        return idx

    def _calculate_filtering_threshold(self, counts, return_plot=False):
        """
        Calculate the filtering thresholds for the given counts.

        Parameters
        ----------
        counts : numpy.ndarray
            The counts of the data.

        Returns
        -------
        tuple
            A tuple containing the lower and upper filtering thresholds.

        Examples
        --------
        >>> import numpy as np
        >>> from sparcscore.processing.filtering import SizeFilter
        >>> np.random.seed(0)
        >>> counts1 = np.random.normal(0, 1, 1000)
        >>> counts2 = np.random.normal(3, 1, 1000)
        >>> counts = np.concatenate((counts1, counts2))
        >>> size_filter = SizeFilter(log=False)
        >>> thresholds = size_filter.calculate_filtering_threshold(counts)
        >>> thresholds
        (1.0785720179538172, 4.911918901671607)
        """

        # take the log of the counts if log_values is True
        if self.log_values:
            data = np.log(counts)
        else:
            data = counts.copy()

        # reshape the data to have the correct dimensions for fitting the model
        data_reshaped = data.reshape(-1, 1)

        # initialize and fit the model
        gmm = GaussianMixture(n_components=self.n_components, random_state=0)
        gmm.fit(data_reshaped)

        # Get the means, variances, and weights of the fitted Gaussians
        means = gmm.means_.flatten()
        variances = gmm.covariances_.flatten()
        weights = gmm.weights_

        # get index of the model which matches to the population of cells that should be kept
        idx = self._get_index_population(means, weights)

        # calculate the thresholds for the selected model using the given confidence interval
        mu = means[idx]
        sigma = np.sqrt(variances[idx])

        percent = 1 - self.confidence_interval
        lower = percent / 2
        upper = 1 - percent / 2

        lower_threshold = mu + sigma * norm.ppf(lower)
        upper_threshold = mu + sigma * norm.ppf(upper)

        threshold = (lower_threshold, upper_threshold)

        fig = self._get_gaussian_model_plot(
            counts=data,
            means=means,
            variances=variances,
            weights=weights,
            threshold=threshold,
        )

        if self.plot_qc:
            plt.show(fig)  # show the figure if plot_qc is True

        if self.log_values:
            self.filter_threshold = np.exp(threshold)
        else:
            self.filter_threshold = threshold

        self.log(
            f"Calculated threshold for {self.label} with {self.confidence_interval * 100}% confidence interval: {self.filter_threshold}"
        )

        if return_plot:
            return fig

    def _get_ids_to_remove(self, input_mask):
        """
        Get the IDs to remove from the input mask based on the filtering threshold.

        Parameters
        ----------
        input_mask : ndarray
            The input mask as a numpy array.

        Returns
        -------
        ndarray
            An array containing the IDs to remove from the input mask.

        Notes
        -----
        This function calculates the filtering threshold based on the pixel counts of the input mask.
        It then identifies the IDs that are outside of the chosen threshold range and returns them as an array.

        The filtering threshold can be automatically calculated if not provided.
        """

        self.load_mask(input_mask)

        # get value counts of the mask
        counts = np.unique(self.mask, return_counts=True)

        self.ids = counts[0][1:]
        pixel_counts = counts[1][1:]

        fig = self._get_histogram_plot(pixel_counts, self.label)
        plt.close(fig)

        if self.plot_qc:
            plt.show(fig)

        # automatically calculate filtering threshold if not provided
        if self.filter_threshold is None:
            self._calculate_filtering_threshold(pixel_counts)

        ids_remove = []

        if self.filter_lower:
            _ids = counts[0][1:][np.where(pixel_counts < self.filter_threshold[0])]
            ids_remove.extend(_ids)

            self.log(
                f"Found {len(_ids)} ids to remove from {self.label} mask which are smaller than the chosen threshold range {self.filter_threshold}."
            )
        else:
            _ids = counts[0][1:][np.where(pixel_counts < self.filter_threshold[0])]
            self.log(
                f"Filtering lower threshold is disabled. Not filtering {len(_ids)} that fall below the threshold range {self.filter_threshold}."
            )

        if self.filter_upper:
            _ids = counts[0][1:][np.where(pixel_counts > self.filter_threshold[1])]
            ids_remove.extend(_ids)

            self.log(
                f"Found {len(_ids)} ids to remove from {self.label} mask which are bigger than the chosen threshold range {self.filter_threshold}."
            )
        else:
            _ids = counts[0][1:][np.where(pixel_counts > self.filter_threshold[1])]
            self.log(
                f"Filtering upper threshold is disabled. Not filtering {len(_ids)} that fall above the threshold range {self.filter_threshold}."
            )

        self.ids_to_remove = ids_remove

    def visualize_filtering_results(self, return_fig=True, return_maps=False, plot_fig=True):
        mask = self.mask.copy()

        class_ids = set(self.ids)

        # get the ids to visualize as red for discarded
        ids_discard = list(self.ids_to_remove)
        ids_keep = list(class_ids - set(self.ids_to_remove))

        # generate masks for the two classes
        final_mask = np.where(np.isin(mask, ids_keep), 2, mask)
        final_mask = np.where(np.isin(mask, ids_discard), 1, final_mask)

        # only plot results if requested
        if not plot_fig:
            if return_maps:
                return final_mask
        else:
            # generate the plot
            cmap, norm = _custom_cmap()

            fig, axs = plt.subplots(1, 1, figsize=(10, 10))

            # visualize masks
            if len(final_mask.shape) == 3:
                axs.imshow(np.zeros(final_mask[0]), cmap=cmap, norm=norm)
                axs.imshow(final_mask[0], cmap=cmap, norm=norm)
            elif len(final_mask.shape) == 2:
                axs.imshow(np.zeros(final_mask), cmap=cmap, norm=norm)
                axs.imshow(final_mask, cmap=cmap, norm=norm)

            axs.axis("off")
            axs.set_title("Filtered {self.label} mask")

            fig.tight_layout()
            if not return_fig:
                plt.show()

                if return_maps:
                    return final_mask
            else:
                if return_maps:
                    return fig, final_mask
                else:
                    return fig

    def filter(self, input_mask):
        """
        Filter the input mask based on the filtering threshold.

        Parameters
        ----------
        input_mask : ndarray
            The input mask to be filtered. Expected shape is (X, Y)

        Returns
        -------
        filtered_mask : ndarray
            The filtered mask after settings the IDs which do not fullfill the filtering criteria to 0.

        """

        if self.ids_to_remove is None:
            self._get_ids_to_remove(input_mask)

        return self.get_updated_mask(input_mask, self.ids_to_remove)


class MatchNucleusCytosolIds(_BaseFilter):
    """
    Filter class for matching nucleusIDs to their matching cytosol IDs and removing all classes from the given
    segmentation masks that do not fullfill the filtering criteria.

    Masks only pass filtering if both a nucleus and a cytosol mask are present and have an overlapping area
    larger than the specified threshold. If the threshold is not specified, the default value is set to 0.5.

    Parameters
    ----------
    filtering_threshold : float, optional
        The threshold for filtering cytosol IDs based on the proportion of overlapping area with the nucleus. Default is 0.5.
    downsampling_factor : int, optional
        The downsampling factor for the masks. Default is None.
    erosion_dilation : bool, optional
        Flag indicating whether to perform erosion and dilation on the masks during upscaling. Default is True.
    smoothing_kernel_size : int, optional
        The size of the smoothing kernel for upscaling. Default is 7.
    *args
        Additional positional arguments.
    **kwargs
        Additional keyword arguments.

    Attributes
    ----------
    filtering_threshold : float
        The threshold for filtering cytosol IDs based on the proportion of overlapping area with the nucleus.
    downsample : bool
        Flag indicating whether downsampling is enabled.
    downsampling_factor : int
        The downsampling factor for the masks.
    erosion_dilation : bool
        Flag indicating whether to perform erosion and dilation on the masks during upscaling.
    smoothing_kernel_size : int
        The size of the smoothing kernel for upscaling.
    nucleus_mask : numpy.ndarray
        The nucleus mask.
    cytosol_mask : numpy.ndarray
        The cytosol mask.
    nuclei_discard_list : list
        A list of nucleus IDs to be discarded.
    cytosol_discard_list : list
        A list of cytosol IDs to be discarded.
    nucleus_lookup_dict : dict
        A dictionary mapping nucleus IDs to matched cytosol IDs after filtering.

    Methods
    -------
    load_masks(nucleus_mask, cytosol_mask)
        Load the nucleus and cytosol masks.
    update_cytosol_mask(cytosol_mask)
        Update the cytosol mask based on the matched nucleus-cytosol pairs.
    update_masks()
        Update the nucleus and cytosol masks after filtering.
    match_nucleus_id(nucleus_id)
        Match the given nucleus ID to a cytosol ID based on the overlapping area.
    initialize_lookup_table()
        Initialize the lookup table by matching all nucleus IDs to cytosol IDs.
    count_cytosol_occurances()
        Count the occurrences of each cytosol ID in the lookup table.
    check_for_unassigned_cytosols()
        Check for unassigned cytosol IDs in the cytosol mask.
    identify_multinucleated_cells()
        Identify and discard multinucleated cells from the lookup table.
    cleanup_filtering_lists()
        Cleanup the discard lists by removing duplicate entries.
    cleanup_lookup_dictionary()
        Cleanup the lookup dictionary by removing discarded nucleus-cytosol pairs.
    generate_lookup_table(nucleus_mask, cytosol_mask)
        Generate the lookup table by performing all necessary steps.
    filter(nucleus_mask, cytosol_mask)
        Filter the nucleus and cytosol masks based on the matching results.

    """

    def __init__(
        self,
        filtering_threshold=0.5,
        downsampling_factor=None,
        erosion_dilation=True,
        smoothing_kernel_size=7,
        directory=None,
        *args,
        **kwargs,
    ):
        if directory is not None:
            self.directory = directory
        else:
            self.directory = os.getcwd()

        super().__init__(*args, directory=self.directory, **kwargs)

        # set relevant parameters
        self.filtering_threshold = filtering_threshold

        # set downsampling parameters
        if downsampling_factor is not None:
            self.downsample = True
            self.downsampling_factor = downsampling_factor
            self.erosion_dilation = erosion_dilation
            self.smoothing_kernel_size = smoothing_kernel_size
        else:
            self.downsample = False

        # initialize placeholders for masks
        self.nucleus_mask = None
        self.cytosol_mask = None

        # initialize datastructures for saving results
        self._nucleus_lookup_dict = {}
        self.nuclei_discard_list = []
        self.cytosol_discard_list = []
        self.nucleus_lookup_dict = None

    def _load_masks(self, nucleus_mask, cytosol_mask):
        """
        Load the nucleus and cytosol masks into their placeholders.
        This function only loads the masks (and downsamples them if necessary) if this has not already been performed.

        Parameters
        ----------
        nucleus_mask : numpy.ndarray
            The nucleus mask.
        cytosol_mask : numpy.ndarray
            The cytosol mask.
        """
        if self.downsample:
            if self.nucleus_mask is None:
                self.nucleus_mask = self.downsample_mask(nucleus_mask)
            if self.cytosol_mask is None:
                self.cytosol_mask = self.downsample_mask(cytosol_mask)
        else:
            if self.nucleus_mask is None:
                self.nucleus_mask = nucleus_mask
            if self.cytosol_mask is None:
                self.cytosol_mask = cytosol_mask

    def _get_updated_cytosol_mask(self, cytosol_mask):
        """
        Update the cytosol mask based on the matched nucleus-cytosol pairs.

        Parameters
        ----------
        cytosol_mask : numpy.ndarray
            The cytosol mask.

        Returns
        -------
        numpy.ndarray
            The updated cytosol mask.
        """
        updated = np.zeros_like(cytosol_mask, dtype=bool)

        # remove cytosols that need to be deleted
        cytosol_mask = self.get_updated_mask(cytosol_mask, self.cytosol_discard_list)

        for nucleus_id, cytosol_id in self.nucleus_lookup_dict.items():
            condition = np.logical_and(cytosol_mask == cytosol_id, ~updated)
            cytosol_mask[condition] = nucleus_id
            updated = np.logical_or(updated, condition)
        return cytosol_mask

    def _get_updated_masks(self):
        """
        Update the nucleus and cytosol masks after filtering.

        Returns
        -------
        tuple
            A tuple containing the updated nucleus mask and cytosol mask.
        """
        nucleus_mask = self.get_updated_mask(self.nucleus_mask, self.nuclei_discard_list)
        cytosol_mask = self._get_updated_cytosol_mask(self.cytosol_mask)

        if self.downsample:
            nucleus_mask = self.get_upscaled_mask_basic(nucleus_mask, self.erosion_dilation)
            cytosol_mask = self.get_upscaled_mask_basic(self.cytosol_mask, self.erosion_dilation)

        return nucleus_mask, cytosol_mask

    def _match_nucleus_id(self, nucleus_id):
        """
        Match the given nucleus ID to a cytosol ID based on the overlapping area.

        Parameters
        ----------
        nucleus_id : int
            The nucleus ID to be matched.

        Returns
        -------
        int or None
            The matched cytosol ID, or None if no match is found.
        """
        nucleus_pixels = np.where(self.nucleus_mask == nucleus_id)
        potential_cytosol = self.cytosol_mask[nucleus_pixels]

        if np.all(potential_cytosol != 0):
            unique_cytosol, counts = np.unique(potential_cytosol, return_counts=True)
            all_counts = np.sum(counts)
            cytosol_proportions = counts / all_counts

            if np.any(cytosol_proportions >= self.filtering_threshold):
                cytosol_id = unique_cytosol[np.argmax(cytosol_proportions >= self.filtering_threshold)]
                if cytosol_id != 0:
                    self._nucleus_lookup_dict[nucleus_id] = cytosol_id
                    return cytosol_id
                else:
                    self.nuclei_discard_list.append(nucleus_id)
                    return None
            else:
                self.nuclei_discard_list.append(nucleus_id)
                return None
        else:
            self.nuclei_discard_list.append(nucleus_id)
            return None

    def _initialize_lookup_table(self):
        """
        Initialize the lookup table by matching all nucleus IDs to cytosol IDs.
        """
        all_nucleus_ids = self.get_unique_ids(self.nucleus_mask)

        for nucleus_id in all_nucleus_ids:
            self._match_nucleus_id(nucleus_id)

    def _count_cytosol_occurances(self):
        """
        Count the occurrences of each cytosol ID in the lookup table.
        """
        cytosol_count = defaultdict(int)

        for cytosol in self._nucleus_lookup_dict.values():
            cytosol_count[cytosol] += 1

        self.cytosol_count = cytosol_count

    def _check_for_unassigned_cytosols(self):
        """
        Check for unassigned cytosol IDs in the cytosol mask.
        """
        all_cytosol_ids = self.get_unique_ids(self.cytosol_mask)

        for cytosol_id in all_cytosol_ids:
            if cytosol_id not in self._nucleus_lookup_dict.values():
                self.cytosol_discard_list.append(cytosol_id)

    def _identify_multinucleated_cells(self):
        """
        Identify and discard multinucleated cells from the lookup table.
        """
        for nucleus, cytosol in self._nucleus_lookup_dict.items():
            if self.cytosol_count[cytosol] > 1:
                self.nuclei_discard_list.append(nucleus)
                self.cytosol_discard_list.append(cytosol)

    def _cleanup_filtering_lists(self):
        """
        Cleanup the discard lists by removing duplicate entries.
        """
        self.nuclei_discard_list = list(set(self.nuclei_discard_list))
        self.cytosol_discard_list = list(set(self.cytosol_discard_list))

    def _cleanup_lookup_dictionary(self):
        """
        Cleanup the lookup dictionary by removing discarded nucleus-cytosol pairs.
        """
        _cleanup = []
        for nucleus_id, cytosol_id in self._nucleus_lookup_dict.items():
            if nucleus_id in self.nuclei_discard_list:
                _cleanup.append(nucleus_id)
            if cytosol_id in self.cytosol_discard_list:
                _cleanup.append(nucleus_id)

        # ensure we have no duplicate entries
        _cleanup = list(set(_cleanup))
        for nucleus in _cleanup:
            del self._nucleus_lookup_dict[nucleus]

    def get_lookup_table(self, nucleus_mask, cytosol_mask):
        """
        Generate the lookup table by performing all necessary steps.

        Parameters
        ----------
        nucleus_mask : numpy.ndarray
            The nucleus mask.
        cytosol_mask : numpy.ndarray
            The cytosol mask.

        Returns
        -------
        dict
            The lookup table mapping nucleus IDs to matched cytosol IDs.
        """
        self._load_masks(nucleus_mask, cytosol_mask)
        self._initialize_lookup_table()
        self._count_cytosol_occurances()
        self._check_for_unassigned_cytosols()
        self._identify_multinucleated_cells()
        self._cleanup_filtering_lists()
        self._cleanup_lookup_dictionary()

        self.nucleus_lookup_dict = self._nucleus_lookup_dict

        return self.nucleus_lookup_dict

    def visualize_filtering_results(self, return_fig=True, return_maps=False, plot_fig=True):
        nuc_mask = self.nucleus_mask.copy()
        cyto_mask = self.cytosol_mask.copy()

        class_ids_nuc = set(np.unique(nuc_mask)) - {0}
        class_ids_cyto = set(np.unique(cyto_mask)) - {0}

        # get the ids to visualize as red for discarded
        ids_discard_nuc = self.nuclei_discard_list
        ids_keep_nuc = list(set(class_ids_nuc) - set(ids_discard_nuc))

        ids_discard_cyto = self.cytosol_discard_list
        ids_keep_cyto = list(set(class_ids_cyto) - set(ids_discard_cyto))

        # generate masks for the two classes
        final_mask_nuc = np.where(np.isin(nuc_mask, ids_keep_nuc), 2, nuc_mask)
        final_mask_nuc = np.where(np.isin(nuc_mask, ids_discard_nuc), 1, final_mask_nuc)

        final_mask_cyto = np.where(np.isin(cyto_mask, ids_keep_cyto), 2, cyto_mask)
        final_mask_cyto = np.where(np.isin(cyto_mask, ids_discard_cyto), 1, final_mask_cyto)

        # only plot results if requested
        if not plot_fig:
            if return_maps:
                return final_mask_nuc, final_mask_cyto
        else:
            # generate the plot
            cmap, norm = _custom_cmap()

            fig, axs = plt.subplots(1, 1, figsize=(10, 10))

            # visualize masks
            if len(final_mask_nuc.shape) == 3:
                axs[0].imshow(final_mask_nuc[0], cmap=cmap, norm=norm)
                axs[0].imshow(final_mask_cyto[0], cmap=cmap, norm=norm)
            elif len(final_mask_nuc.shape) == 2:
                axs[0].imshow(final_mask_nuc, cmap=cmap, norm=norm)
                axs[0].imshow(final_mask_cyto, cmap=cmap, norm=norm)

            axs[0].axis("off")
            axs[0].set_title("Filtered Nucleus and Cytosol Masks")

            fig.tight_layout()
            if not return_fig:
                plt.show()

                if return_maps:
                    return final_mask_nuc, final_mask_cyto
            else:
                if return_maps:
                    return fig, final_mask_nuc, final_mask_cyto
                else:
                    return fig

    def filter(self, nucleus_mask, cytosol_mask):
        """
        Filter the nucleus and cytosol masks based on the matching results and return the updated masks.

        Parameters
        ----------
        nucleus_mask : numpy.ndarray
            The nucleus mask.
        cytosol_mask : numpy.ndarray
            The cytosol mask.

        Returns
        -------
        tuple
            A tuple containing the updated nucleus mask and cytosol mask after filtering.
        """
        self._load_masks(nucleus_mask, cytosol_mask)

        if self.nucleus_lookup_dict is None:
            self.get_lookup_table(nucleus_mask, cytosol_mask)

        return self._get_updated_masks()



import os
import sys
import time
import timeit

import matplotlib.pyplot as plt
import numpy as np
import xarray
from skimage.filters import median
from skimage.morphology import dilation, disk, erosion

from scportrait.pipeline.segmentation.segmentation import (
    Segmentation,
)
from scportrait.plotting._utils import _custom_cmap
from scportrait.processing.images._image_processing import downsample_img, percentile_normalization
from scportrait.processing.masks.mask_filtering import MatchNucleusCytosolIds, SizeFilter


class _BaseSegmentation(Segmentation):
    """Base Segmentation Workflow class that implements methods that can be reused across different segmentation workflows."""

    MASK_NAMES = ["nucleus", "cytosol"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_channel_selection()
        self.nGPUs = None

    def _setup_channel_selection(self):
        self._setup_maximum_intensity_projection()
        self._define_channels_to_extract_for_segmentation()
        self._remap_maximum_intensity_projection_channels()

    def _setup_maximum_intensity_projection(self):
        # check if channels that should be maximum intensity projected and combined are defined in the config
        if "combine_cytosol_channels" in self.config.keys():
            self.combine_cytosol_channels = self.config["combine_cytosol_channels"]
            assert isinstance(
                self.combine_cytosol_channels, list
            ), "combine_cytosol_channels must be a list of integers specifying the indexes of the channels to combine."
            assert (
                len(self.combine_cytosol_channels) > 1
            ), "combine_cytosol_channels must contain at least two integers specifying the indexes of the channels to combine."
            self.maximum_project_cytosol = True
        else:
            self.combine_cytosol_channels = None
            self.maximum_project_cytosol = False

        if "combine_nucleus_channels" in self.config.keys():
            self.combine_nucleus_channels = self.config["combine_nucleus_channels"]
            assert isinstance(
                self.combine_nucleus_channels, list
            ), "combine_nucleus_channels must be a list of integers specifying the indexes of the channels to combine."
            assert (
                len(self.combine_nucleus_channels) > 1
            ), "combine_nucleus_channels must contain at least two integers specifying the indexes of the channels to combine."
            self.maximum_project_nucleus = True
        else:
            self.combine_nucleus_channels = None
            self.maximum_project_nucleus = False

    def _define_channels_to_extract_for_segmentation(self):
        self.segmentation_channels = []

        if "nucleus" in self.MASK_NAMES:
            if "segmentation_channel_nuclei" in self.config.keys():
                self.nucleus_segmentation_channel = self.config["segmentation_channel_nuclei"]
            elif "combine_nucleus_channels" in self.config.keys():
                self.nucleus_segmentation_channel = self.combine_nucleus_channels
            else:
                self.nucleus_segmentation_channel = self.DEFAULT_NUCLEI_CHANNEL_IDS

            self.segmentation_channels.extend(self.nucleus_segmentation_channel)

        if "cytosol" in self.MASK_NAMES:
            if "segmentation_channel_cytosol" in self.config.keys():
                self.cytosol_segmentation_channel = self.config["segmentation_channel_cytosol"]
            elif "combine_cytosol_channels" in self.config.keys():
                self.cytosol_segmentation_channel = self.combine_cytosol_channels
            else:
                self.cytosol_segmentation_channel = self.DEFAULT_CYTOSOL_CHANNEL_IDS

            self.segmentation_channels.extend(self.cytosol_segmentation_channel)

        # remove any duplicate entries and sort according to order
        self.segmentation_channels = list(set(self.segmentation_channels))

        # check validity of resulting list of segmentation channels
        assert len(self.segmentation_channels) > 0, "No segmentation channels specified in config file."
        assert (
            len(self.segmentation_channels) >= self.N_INPUT_CHANNELS
        ), f"Fewer segmentation channels {self.segmentation_channels} provided than expected by segmentation method {self.N_INPUT_CHANNELS}."

        if len(self.segmentation_channels) > self.N_INPUT_CHANNELS:
            assert (
                self.maximum_project_nucleus or self.maximum_project_cytosol
            ), "More input channels provided than accepted by the segmentation method and no maximum intensity projection performed on any of the input values."

    def _remap_maximum_intensity_projection_channels(self):
        """After selecting channels that are passed to the segmentation update indexes of the channels for maximum intensity projection so that they reflect the provided image subset"""
        if self.maximum_project_nucleus:
            self.original_combine_nucleus_channels = self.combine_nucleus_channels
            self.combine_nucleus_channels = [self.segmentation_channels.index(x) for x in self.combine_nucleus_channels]
        if self.maximum_project_cytosol:
            self.original_combine_cytosol_channels = self.combine_cytosol_channels
            self.combine_cytosol_channels = [self.segmentation_channels.index(x) for x in self.combine_cytosol_channels]

    def _transform_input_image(self, input_image):
        start_transform = timeit.default_timer()

        if isinstance(input_image, xarray.DataArray):
            input_image = input_image.data.compute()

        values = []
        # check if any channels need to be transformed
        if "nucleus" in self.MASK_NAMES:
            if self.maximum_project_nucleus:
                self.log(
                    f"For nucleus segmentation using the maximum intensity projection of channels {self.original_combine_nucleus_channels}."
                )
                nucleus_channel = self._maximum_project_channels(input_image, self.combine_nucleus_channels)
                if len(nucleus_channel.shape) == 2:
                    nucleus_channel = nucleus_channel[np.newaxis, ...]
                if len(nucleus_channel.shape) == 4:
                    nucleus_channel = nucleus_channel.squeeze()
            else:
                nucleus_channel = input_image[self.DEFAULT_NUCLEI_CHANNEL_IDS]
                if len(nucleus_channel.shape) == 2:
                    nucleus_channel = nucleus_channel[np.newaxis, ...]
                if len(nucleus_channel.shape) == 4:
                    nucleus_channel = nucleus_channel.squeeze()
            values.append(nucleus_channel)

        if "cytosol" in self.MASK_NAMES:
            if self.maximum_project_cytosol:
                self.log(
                    f"For cytosol segmentation using the maximum intensity projection of channels {self.original_combine_cytosol_channels}."
                )
                cytosol_channel = self._maximum_project_channels(input_image, self.combine_cytosol_channels)
                if len(cytosol_channel.shape) == 4:
                    cytosol_channel = cytosol_channel.squeeze()
                if len(cytosol_channel.shape) == 2:
                    cytosol_channel = cytosol_channel[np.newaxis, ...]
            else:
                cytosol_channel = input_image[self.DEFAULT_CYTOSOL_CHANNEL_IDS]
                if len(cytosol_channel.shape) == 4:
                    cytosol_channel = cytosol_channel.squeeze()
                if len(cytosol_channel.shape) == 2:
                    cytosol_channel = cytosol_channel[np.newaxis, ...]

            values.append(cytosol_channel)

        input_image = np.vstack(values)

        assert (
            input_image.shape[0] == self.N_INPUT_CHANNELS
        ), f"Number of channels in input image {input_image.shape[0]} does not match the number of channels expected by segmentation method {self.N_INPUT_CHANNELS}."

        stop_transform = timeit.default_timer()
        self.transform_time = stop_transform - start_transform
        return input_image

    def return_empty_mask(self, input_image):
        _, x, y = input_image.shape
        self._save_segmentation_sdata(np.zeros((self.N_MASKS, x, y)), [])

    def _check_seg_dtype(self, mask: np.ndarray, mask_name: str) -> np.ndarray:
        if not isinstance(mask, self.DEFAULT_SEGMENTATION_DTYPE):
            Warning(
                f"{mask_name} segmentation map is not of the correct dtype. \n Forcefully converting {mask.dtype} to {self.DEFAULT_SEGMENTATION_DTYPE}. \n This could lead to unexpected behaviour."
            )

            return mask.astype(self.DEFAULT_SEGMENTATION_DTYPE)

        else:
            return mask

    #### Downsampling ####
    def _get_downsampling_parameters(self) -> None:
        self.N = self.config["downsampling_factor"]

        if "smoothing_kernel_size" in self.config.keys():
            self.smoothing_kernel_size = self.config["smoothing_kernel_size"]

            if self.smoothing_kernel_size > self.N:
                self.log(
                    "Warning: Smoothing Kernel size is larger than the downsampling factor. This can lead to issues during smoothing where segmentation masks are lost. Please ensure to double check your results."
                )

        else:
            self.log(
                "Smoothing Kernel size not explicitly defined. Will calculate a default value based on the downsampling factor."
            )
            self.smoothing_kernel_size = self.N

        return None

    def _calculate_padded_image_size(self, img: np.array) -> None:
        """prepare metrics for image downsampling. Calculates image padding required for downsampling and returns
        metrics for this as well as resulting downsampled image size.
        """

        self.input_image_size = img.shape

        # check if N fits perfectly into image shape if not calculate how much we need to pad
        _, x, y = self.input_image_size

        if x % self.N == 0:
            pad_x = (0, 0)
        else:
            pad_x = (0, self.N - x % self.N)

        if y % self.N == 0:
            pad_y = (0, 0)
        else:
            pad_y = (0, self.N - y % self.N)

        # calculate resulting image size for use when e.g. inititalizing empty arrays to save results to
        padded_image_size = (
            2,
            self.input_image_size[1] + pad_x[1],
            self.input_image_size[2] + pad_y[1],
        )

        self.expected_padded_image_size = padded_image_size
        self.pad_x = pad_x
        self.pad_y = pad_y

        # track original image size
        self.original_image_size = img.shape

        return None

    def _downsample_image(self, img: np.array, debug: bool = False) -> np.array:
        """
        Downsample image by a factor of N. Before downsampling this function will pad the image to ensure its compatible with the selected kernel size.

        Parameters
        ----------
        img
            image to be downsampled

        Returns
        -------
        downsampled image

        """

        self.log(
            f"Performing image padding to ensure that image is compatible with selected downsample kernel size of {self.N}."
        )

        # perform image padding to ensure that image is compatible with downsample kernel size
        img = np.pad(img, ((0, 0), self.pad_x, self.pad_y))
        self.padded_image_size = img.shape

        if debug:
            self.log(
                f"Original image had size {self.expected_padded_image_size}, padded image is {self.padded_image_size}"
            )

        # sanity check to make sure padding worked as we wanted
        if self.expected_padded_image_size != self.padded_image_size:
            Warning(
                f"Expected a padded image of size {self.expected_padded_image_size} but got {self.padded_image_size}. Padding did not work as expted"
            )
            sys.exit(
                f"Error. Image padding did not work as expected. Expected a padded image of size {self.expected_padded_image_size} but got {self.padded_image_size}."
            )

        self.log(f"Downsampling image by a factor of {self.N}x{self.N}")

        # actually perform downsampling
        img = downsample_img(img, N=self.N)

        self.downsampled_image_size = img.shape

        if debug:
            self.log(f"Downsampled image size {self.downsampled_image_size}")

        return img

    def _rescale_downsampled_mask(self, mask: np.array, mask_name: str) -> np.array:
        input_mask = mask.copy()

        # get number of objects in mask for sanity checking
        n_classes = len(np.unique(mask))

        # rescale segmentations masks to padded image size
        mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)

        # perform erosion and dilation for smoothing
        mask = erosion(mask, footprint=disk(self.smoothing_kernel_size))
        mask = dilation(
            mask, footprint=disk(self.smoothing_kernel_size + 1)
        )  # dilate 1 more than eroded to ensure that we do not lose any pixels

        # sanity check to make sure that smoothing does not remove masks
        if len(np.unique(mask)) != n_classes:
            Warning(
                "Number of objects in segmentation mask changed after smoothing. This should not happen. Ensure that you have chosen adequate smoothing parameters."
            )

            self.log(
                f"Will recalculate upsampling of {mask_name} mask with lower smoothing value to prevent the number of segmented objects from changing. Please ensure to double check your results."
            )

            smoothing_kernel_size = self.smoothing_kernel_size

            while len(np.unique(mask)) != n_classes:
                smoothing_kernel_size = smoothing_kernel_size - 1

                if smoothing_kernel_size == 0:
                    # if we reach 0 then we do not perform any smoothing
                    # repeat rescaling of the original mask

                    mask = input_mask
                    mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)
                    self.log(f"Did not perform smoothing of {mask_name} mask.")

                    break

                else:
                    mask = input_mask
                    mask = mask.repeat(self.N, axis=0).repeat(self.N, axis=1)

                    # perform erosion and dilation for smoothing
                    mask = erosion(mask, footprint=disk(smoothing_kernel_size))

                    mask = dilation(
                        mask, footprint=disk(smoothing_kernel_size + 1)
                    )  # dilate 1 more than eroded to ensure that we do not lose any pixels

            self.log(
                f"Recalculation of {mask_name} mask successful with smoothing kernel size of {smoothing_kernel_size}."
            )

        # remove padding from mask
        x_trim = self.padded_image_size[1] - self.original_image_size[1]
        y_trim = self.padded_image_size[2] - self.original_image_size[2]

        # sanity check to ensure that we are removing what we addded
        assert x_trim == self.pad_x[1]
        assert y_trim == self.pad_y[1]

        # actually perform trimming

        if len(mask.shape) == 3:
            if x_trim > 0:
                if y_trim > 0:
                    mask = mask[:, :-x_trim, :-y_trim]
                else:
                    mask = mask[:, :-x_trim, :]
            else:
                if y_trim > 0:
                    mask = mask[:, :, :-y_trim]
                else:
                    mask = mask
        elif len(mask.shape) == 2:
            if x_trim > 0:
                if y_trim > 0:
                    mask = mask[:-x_trim, :-y_trim]
                else:
                    mask = mask[:-x_trim, :]
            else:
                if y_trim > 0:
                    mask = mask[:, :-y_trim]
                else:
                    mask = mask

        # check that mask has the correct shape and matches to input image

        if len(mask.shape) == 2:
            assert (mask.shape[0] == self.original_image_size[1]) and (mask.shape[1] == self.original_image_size[2])
        elif len(mask.shape) == 3:
            assert (mask.shape[1] == self.original_image_size[1]) and (mask.shape[2] == self.original_image_size[2])

        return mask

    #### Image Processing #####

    def _maximum_project_channels(self, input_image: np.array, channel_ids: list[int]) -> np.array:
        """add multiple channels together to generate a new channel for segmentation using maximum intensity projection"""

        dtype = input_image.dtype
        new_channel = np.zeros_like(input_image[0], dtype=dtype)

        for channel_id in channel_ids:
            new_channel = np.maximum(new_channel, input_image[channel_id])

        return new_channel

    def _normalize_image(
        self,
        input_image: np.array,
        lower: float | list,
        upper: float | list,
        debug: bool = False,
    ) -> np.array:
        if isinstance(lower, float) and isinstance(upper, float):
            self.log("Normalizing each channel to the same range")
            norm_image = percentile_normalization(input_image, lower, upper)

        elif isinstance(lower, list) and isinstance(upper, list):
            norm_image = []

            for i in range(input_image.shape[0]):
                _lower = lower[i]
                _upper = upper[i]  # type: ignore

                norm_image.append(percentile_normalization(input_image[i], _lower, _upper))

            norm_image = np.array(norm_image)
        else:
            raise ValueError(
                "Lower and upper quantile normalization values must be either floats or dictionary of floats."
            )
        if debug:
            if len(norm_image.shape) == 2:
                fig, ax = plt.subplots(1, 1, figsize=(10, 10))
                ax.imshow(norm_image)
                ax.axis("off")
                ax.set_title("Normalized Image")
            else:
                # visualize output if debug is turned on
                n_channels = norm_image.shape[0]
                fig, axs = plt.subplots(1, n_channels, figsize=(10 * n_channels, 10))
                for i in range(n_channels):
                    axs[i].imshow(norm_image[i])
                    axs[i].axis("off")
                    axs[i].set_title(f"Normalized Channel {i}")

            fig.tight_layout()
            fig_path = os.path.join(self.directory, "normalized_input.png")
            fig.savefig(fig_path)

        return norm_image

    def _median_correct_image(self, input_image, median_filter_size: int, debug: bool = False):
        self.log("Performing median filtering on input image")

        for i in range(input_image.shape[0]):
            input_image[i] = median(input_image[i], disk(median_filter_size))

        if debug:
            if len(input_image.shape) == 2:
                fig, ax = plt.subplots(1, 1, figsize=(10, 10))
                ax.imshow(input_image)
                ax.axis("off")
                ax.set_title("Median Corrected Image")
            else:
                # visualize output if debug is turned on
                n_channels = input_image.shape[0]

                fig, axs = plt.subplots(1, n_channels, figsize=(10 * n_channels, 10))
                for i in range(n_channels):
                    axs[i].imshow(input_image[i])
                    axs[i].axis("off")
                    axs[i].set_title(f"Median Corrected Channel {i}")

            fig.tight_layout()
            fig_path = os.path.join(self.directory, "median_corrected_input.png")
            fig.savefig(fig_path)

        return input_image

    ##### Filtering Functions #####

    # 1. Size Filtering
    def _check_for_size_filtering(self, mask_types: list[str]) -> None:
        """
        Check if size filtering should be performed on the masks.
        If size filtering is turned on, the thresholds for filtering are loaded from the config file.
        """

        assert all(
            mask_type in self.MASK_NAMES for mask_type in mask_types
        ), f"mask_types must be a list of strings that are valid mask names {self.MASK_NAMES}."

        if "filter_masks_size" in self.config.keys():
            self.filter_size = self.config["filter_masks_size"]
        else:
            # default behaviour is this is turned off filtering can always be performed later and this preserves the whole segmentation mask
            self.filter_size = False
            for mask_type in mask_types:
                # save attributes for use later
                setattr(self, f"{mask_type}_thresholds", None)
                setattr(self, f"{mask_type}_confidence_interval", None)

        # load parameters for cellsize filtering
        if self.filter_size:
            for mask_type in mask_types:
                thresholds, confidence_interval = self._get_params_cellsize_filtering(type=mask_type)

                # save attributes for use later
                setattr(self, f"{mask_type}_thresholds", thresholds)
                setattr(self, f"{mask_type}_confidence_interval", confidence_interval)

    def _get_params_cellsize_filtering(self, type) -> tuple[tuple[float] | None, float | None]:
        absolute_filter_status = False

        if "min_size" in self.config[f"{type}_segmentation"].keys():
            min_size = self.config[f"{type}_segmentation"]["min_size"]
            absolute_filter_status = True
        else:
            min_size = None

        if "max_size" in self.config[f"{type}_segmentation"].keys():
            max_size = self.config[f"{type}_segmentation"]["max_size"]
            absolute_filter_status = True
        else:
            max_size = None

        if absolute_filter_status:
            thresholds = [min_size, max_size]
            return (thresholds, None)  # type: ignore
        else:
            thresholds = None

            # get confidence intervals to automatically calculate thresholds
            if "confidence_interval" in self.config[f"{type}_segmentation"].keys():
                confidence_interval = self.config[f"{type}_segmentation"]["confidence_interval"]
            else:
                # get default value
                self.log(f"No confidence interval specified for {type} mask filtering, using default value of 0.95")
                confidence_interval = 0.95

            return (thresholds, confidence_interval)

    def _perform_size_filtering(
        self,
        mask: np.array,
        thresholds: tuple[float, float] | None,
        confidence_interval: float,
        mask_name: str,
        log: bool = True,
        debug: bool = False,
        input_image: np.array = None,
    ) -> np.array:
        """
        Remove elements from mask based on a size filter.

        Parameters
        ----------
        mask
            mask to be filtered
        """
        start_time = time.time()

        if thresholds is not None:
            self.log(f"Performing filtering of {mask_name} with specified thresholds {thresholds} from config file.")
        else:
            self.log(
                f"Automatically calculating thresholds for filtering of {mask_name} based on a fitted normal distribution with a confidence interval of {confidence_interval * 100}%."
            )

        filter = SizeFilter(
            label=mask_name,
            log=log,
            plot_qc=self.debug,
            directory=self.directory,
            confidence_interval=confidence_interval,
            filter_threshold=thresholds,
        )

        filtered_mask = filter.filter(mask)
        self.log(
            f"Removed {len(filter.ids_to_remove)} nuclei as they fell outside of the threshold range {filter.filter_threshold}."
        )

        if self.debug:
            # get visualization of the filtering results (2 = filtered, 1 = keep, 0 = background)
            mask = filter.visualize_filtering_results(plot_fig=False, return_maps=True)

            if not self.is_shard:
                if self.save_filter_results:
                    if len(mask.shape) == 2:
                        # add results to sdata object
                        self.filehandler._write_segmentation_sdata(
                            mask,
                            segmentation_label=f"debugging_seg_size_filter_results_{mask_name}",
                        )
                    else:
                        self.filehandler._write_segmentation_sdata(
                            mask[0],
                            segmentation_label=f"debugging_seg_size_filter_results_{mask_name}",
                        )
                    # then this does not need to be plotted as it can be visualized from there
                    plot_results = False
                else:
                    plot_results = True
            else:
                plot_results = True

            if plot_results:
                # get input image for visualization
                if input_image is not None:
                    if len(input_image.shape) == 2:
                        image_map = input_image
                    else:
                        if mask_name == "nucleus":
                            image_map = input_image[0]
                        elif mask_name == "cytosol":
                            image_map = input_image[1]

                    cmap, norm = _custom_cmap()

                    fig, axs = plt.subplots(1, 1, figsize=(10, 10))
                    axs.imshow(image_map, cmap="gray")
                    axs.imshow(mask, cmap=cmap, norm=norm)
                    axs.axis("off")
                    axs.set_title(f"Visualization of classes removed during {mask_name} size filtering")
                    fig_path = os.path.join(self.directory, f"Results_{mask_name}_size_filtering.png")
                    fig.savefig(fig_path)

                    self._clear_cache(vars_to_delete=[fig, input_image, cmap, norm])

            # clearup memory
            self._clear_cache(vars_to_delete=[mask, plot_results])

        end_time = time.time()

        self.log(f"Total time to perform {mask_name} size filtering: {end_time - start_time} seconds")

        return filtered_mask

    # 2. matching masks
    def _check_for_mask_matching_filtering(self) -> None:
        """Check to see if the masks should be filtered for matching nuclei/cytosols within the segmentation run."""

        DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING = 0.95
        # check to see if the cells should be filtered for matching nuclei/cytosols within the segmentation run
        if "match_masks" in self.config.keys():
            self.filter_match_masks = self.config["match_masks"]
            if "filtering_threshold_mask_matching" in self.config.keys():
                self.mask_matching_filtering_threshold = self.config["filtering_threshold_mask_matching"]
            else:
                self.mask_matching_filtering_threshold = (
                    DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING  # set default parameter
                )

        else:
            # default behaviour that this filtering should be performed, otherwise another additional step is required before extraction
            self.filter_match_masks = True
            self.mask_matching_filtering_threshold = DEFAULT_FILTERING_THRESHOLD_MASK_MATCHING

        # sanity check provided values
        assert isinstance(self.filter_match_masks, bool), "`match_masks` must be a boolean value."
        if self.filter_match_masks:
            assert isinstance(
                self.mask_matching_filtering_threshold, float
            ), "`filtering_threshold_mask_matching` for mask matching must be a float."

    def _perform_mask_matching_filtering(
        self,
        nucleus_mask: np.array,
        cytosol_mask: np.array,
        filtering_threshold: float,
        debug: bool = False,
        input_image: np.array = None,
    ) -> tuple[np.array, np.array]:
        """
        Match the nuclei and cytosol masks to ensure that the same cells are present in both masks.

        Parameters
        ----------
        nucleus_mask
            nucleus mask to be matched
        cytosol_mask
            cytosol mask to be matched
        """
        start_time = time.time()
        self.log("Performing filtering to match Cytosol and Nucleus IDs.")

        # perform filtering to remove cytosols which do not have a corresponding nucleus
        filter = MatchNucleusCytosolIds(filtering_threshold=filtering_threshold)
        masks_nucleus, masks_cytosol = filter.filter(nucleus_mask=nucleus_mask, cytosol_mask=cytosol_mask)

        self.log(
            f"Removed {len(filter.nuclei_discard_list)} nuclei and {len(filter.cytosol_discard_list)} cytosols due to filtering."
        )
        self.log(f"After filtering, {len(filter.nucleus_lookup_dict)} matching nuclei and cytosol masks remain.")

        if debug:
            mask_nuc, mask_cyto = filter.visualize_filtering_results(plot_fig=False, return_maps=True)

            # check if image should be added to sdata or plotted and saved
            if not self.is_shard:
                if self.save_filter_results:
                    # add filtering results to sdata object
                    self.filehandler._write_segmentation_sdata(
                        mask_nuc[0], segmentation_label="debugging_seg_match_mask_results_nucleus"
                    )
                    self.filehandler._write_segmentation_sdata(
                        mask_cyto[0], segmentation_label="debugging_seg_match_mask_result_cytosol"
                    )

                    # then no plotting needs to be performed as the results can be viewed in the sdata object
                    plot_results = False
                else:
                    plot_results = True
            else:
                plot_results = True

            if plot_results:
                if input_image is not None:
                    # convert input image from uint16 to uint8
                    input_image = (input_image / 256).astype(np.uint8)

                    cmap, norm = _custom_cmap()

                    fig, axs = plt.subplots(1, 2, figsize=(20, 10))

                    axs[0].imshow(input_image[0], cmap="gray")
                    axs[0].imshow(mask_nuc[0], cmap=cmap, norm=norm)
                    axs[0].imshow(mask_cyto[0], cmap=cmap, norm=norm)
                    axs[0].axis("off")
                    axs[0].set_title("results overlayed nucleus channel")

                    axs[1].imshow(input_image[1], cmap="gray")
                    axs[1].imshow(mask_nuc[0], cmap=cmap, norm=norm)
                    axs[1].imshow(mask_cyto[0], cmap=cmap, norm=norm)
                    axs[1].axis("off")
                    axs[1].set_title("results overlayed cytosol channel")

                    fig.tight_layout()
                    fig_path = os.path.join(self.directory, "Results_mask_matching.png")
                    fig.savefig(fig_path)

                    self._clear_cache(vars_to_delete=[fig, input_image, cmap, norm])

            # clearup memory
            self._clear_cache(vars_to_delete=[mask_nuc, mask_cyto, plot_results])

        self.log(
            f"Total time to perform nucleus and cytosol mask matching filtering: {time.time() - start_time:.2f} seconds"
        )

        return masks_nucleus, masks_cytosol



import os

import numpy as np
import skimage.exposure
from ashlar.filepattern import FilePatternReader
from ashlar.reg import BioformatsMetadata, BioformatsReader
from skimage.filters import gaussian
from skimage.util import invert

from scportrait.processing.images._image_processing import rescale_image


class FilePatternReaderRescale(FilePatternReader):
    """Class for reading images based on a file pattern. If desired the images can be rescaled to a certain range while reading."""

    def __init__(
        self,
        path: str,
        pattern: str,
        overlap: float,
        pixel_size: int = 1,
        do_rescale: bool = False,
        no_rescale_channel: str = "Alexa488",
        rescale_range: tuple[int, int] = (1, 99),
        dtype=None,
    ):
        try:
            super().__init__(path, pattern, overlap, pixel_size=pixel_size)
        except (OSError, FileNotFoundError):
            print(
                f"Error: Could not read images with the given pattern {pattern}. Please check the path {path} and pattern."
            )
            found_files = os.listdir(path)
            print(
                f"At the provided location the the files follow the naming convention:{found_files[0:max(5, len(found_files))]} "
            )

        self.do_rescale = do_rescale
        self.no_rescale_channel = no_rescale_channel
        self.rescale_range = rescale_range
        self.dtype_image = dtype

    @staticmethod
    def rescale(
        img: np.ndarray, rescale_range: tuple[int, int] = (1, 99), cutoff_threshold: int | None = None
    ) -> np.ndarray:
        return rescale_image(img, rescale_range, cutoff_threshold=cutoff_threshold)

    @staticmethod
    # placeholer method kept for compatibility with old code
    # should be reimplemented in the future to allow for more flexible illumination correction
    def correct_illumination(
        img: np.ndarray,
        sigma: float = 30,
        double_correct: bool = False,
        rescale_range: tuple[int, int] = (1, 99),
        cutoff_threshold: int | None = None,
    ) -> np.ndarray:
        img = rescale_image(img, rescale_range, cutoff_threshold=cutoff_threshold, return_float=True)

        # calculate correction mask
        correction = gaussian(img, sigma)
        correction = invert(correction)
        correction = skimage.exposure.rescale_intensity(correction, out_range=(0, 1))

        correction_lows = np.where(img > 0.5, 0, img) * correction
        img_corrected = skimage.exposure.rescale_intensity(img + correction_lows, out_range=(0, 1))

        if double_correct:
            correction_mask_highs = invert(correction)
            correction_mask_highs_02 = skimage.exposure.rescale_intensity(
                np.where(img_corrected < 0.5, 0, img_corrected) * correction_mask_highs
            )
            img_corrected_double = skimage.exposure.rescale_intensity(img_corrected - 0.25 * correction_mask_highs_02)

            return (img_corrected_double * 65535).astype("uint16")
        else:
            return (img_corrected * 65535).astype("uint16")

    def read(self, series, c) -> np.ndarray:
        img = super().read(series, c)

        # check to ensure we are always using the same dtype
        if self.dtype_image is not None:
            if not isinstance(img.dtype, self.dtype_image):
                Warning(f"Found image with dtype {img.dtype}. Automatically converting to {self.dtype_image}")
                img = img.astype(self.dtype_image)
        else:
            # first time reading an image set dtype if none is supplied
            print("Setting dtype to", img.dtype)
            self.dtype_image = type(img.dtype)

        if not self.do_rescale:
            return img
        else:
            if c not in self.no_rescale_channel:
                # get rescale_range for channel c
                if isinstance(self.rescale_range, dict):
                    rescale_range = self.rescale_range[c]
                else:
                    rescale_range = self.rescale_range

                # actually read the image with rescaling applied
                return self.rescale(img, rescale_range=rescale_range)
            else:
                return img


class BioformatsMetadataRescale(BioformatsMetadata):
    """Reimplementation of BioformatsMetadata class to provide same parameters as contained in FilePatternReaderRescale"""

    def __init__(self, path):
        super().__init__(path)

    @property
    def channel_map(self):
        n_channels = self._metadata.getChannelCount(0)
        channel_names = []

        for id in range(n_channels):
            channel_names.append(self._metadata.getChannelName(0, id))
        channel_map = dict(zip(list(range(n_channels)), channel_names, strict=False))

        return channel_map


class BioformatsReaderRescale(BioformatsReader):
    """Class for reading images from Bioformats files (e.g. nd2).

    If desired the images can be rescaled to a certain range while reading.
    """

    def __init__(self, path, plate=None, well=None, do_rescale=False, no_rescale_channel=None, rescale_range=(1, 99)):
        self.path = path
        self.metadata = BioformatsMetadataRescale(self.path)
        self.metadata.set_active_plate_well(plate, well)
        self.do_rescale = do_rescale
        self.no_rescale_channel = no_rescale_channel
        self.rescale_range = rescale_range

    @staticmethod
    def rescale(img, rescale_range=(1, 99)):
        img = skimage.util.img_as_float32(img)
        cutoff1, cutoff2 = rescale_range

        if img.max() > (40000 / 65535):
            _img = img.copy()
            _img[_img > (10000 / 65535)] = 0
            p1 = np.percentile(_img, cutoff1)
            p99 = np.percentile(_img, cutoff2)
        else:
            p1 = np.percentile(img, cutoff1)
            p99 = np.percentile(img, cutoff2)

        img = skimage.exposure.rescale_intensity(img, in_range=(p1, p99), out_range=(0, 1))
        return (img * 65535).astype("uint16")

    def read(self, series, c):
        self.metadata._reader.setSeries(self.metadata.active_series[series])
        index = self.metadata._reader.getIndex(0, c, 0)
        byte_array = self.metadata._reader.openBytes(index)
        dtype = self.metadata.pixel_dtype  # no check required here as dtype has to be the same for all
        shape = self.metadata.tile_size(series)
        img = np.frombuffer(byte_array.tostring(), dtype=dtype).reshape(shape)

        if not self.do_rescale:
            return img
        else:
            if c not in self.no_rescale_channel:
                # get rescale_range for channel c
                if isinstance(self.rescale_range, dict):
                    rescale_range = self.rescale_range[c]
                else:
                    rescale_range = self.rescale_range

                # actually read the image with rescaling applied
                return self.rescale(img, rescale_range=rescale_range)
            else:
                return img



import copy
import warnings

import spatialdata


def get_bounding_box_sdata(
    sdata: spatialdata, max_width: int, center_x: int, center_y: int, drop_points: bool = True
) -> spatialdata:
    """apply bounding box to sdata object

    Args:
        sdata: spatialdata object
        max_width: maximum width of the bounding box
        center_x: x coordinate of the center of the bounding box
        center_y: y coordinate of the center of the bounding box

    Returns:
        spatialdata: spatialdata object with bounding box applied
    """
    _sdata = sdata
    # remove points object to improve subsetting
    if drop_points:
        points_keys = list(_sdata.points.keys())
        if len(points_keys) > 0:
            # add check to make sure we aren't deleting a points object that is only in memory
            in_memory_only, _ = _sdata._symmetric_difference_with_zarr_store()
            in_memory_only = [x.split("/")[-1] for x in in_memory_only]

            for x in points_keys:
                if x not in in_memory_only:
                    del _sdata.points[x]
                else:
                    warnings.warn(
                        f"Points object {x} is in memory only and will not be deleted despite the drop_points flag being set to True.",
                        stacklevel=2,
                    )

    width = max_width // 2

    # ensure that the image is large enough
    if center_x - width < 0:
        center_x = width
    if center_y - width < 0:
        center_y = width

    # subset spatialdata object if its too large
    _sdata = _sdata.query.bounding_box(
        axes=["x", "y"],
        min_coordinate=[center_x - width, center_y - width],
        max_coordinate=[center_x + width, center_y + width],
        target_coordinate_system="global",
    )

    if drop_points:
        # re-add points object
        __sdata = spatialdata.SpatialData.read(sdata.path, selection=["points"])
        for x in points_keys:
            sdata[x] = __sdata[x]
        del __sdata

    return _sdata



"""
sdata
=======

Functions to work with spatialdata objects.
"""

# has to be done at the end, after everything has been imported
import sys

from . import processing as pp
from .write._helper import add_element_sdata, rename_image_element
from .write._write import image as write_image
from .write._write import labels as write_labels

# update symlinks to the functions
sys.modules.update(
    {
        f"{__name__}.{m}": globals()[m]
        for m in [
            "pp",
            "write_image",
        ]
    }
)

__all__ = ["write_image", "write_labels", "add_element_sdata", "rename_image_element", "pp"]



import numpy as np
from mahotas import sobel


def EDF(image: np.ndarray) -> np.ndarray:
    """Calculate Extended Depth of Field for the given input image Z-stack.
    Based on implementation here: https://mahotas.readthedocs.io/en/latest/edf.html#id3

    Args:
        image: Input image array of shape (Z, X, Y)

    Returns:
        np.array: EDF selected image
    """

    # get image stack sizes
    stack, h, w = image.shape

    # determine in focusness for each pixel
    focus = np.array([sobel(z, just_filter=True) for z in image])

    # select best focal plane for each pixel
    best = np.argmax(focus, 0)

    image = image.reshape((stack, -1))  # image is now (stack, nr_pixels)
    image = image.transpose()  # image is now (nr_pixels, stack)
    r = image[np.arange(len(image)), best.ravel()]  # Select the right pixel at each location
    r = r.reshape((h, w))  # reshape to get final result

    return r


def maximum_intensity_projection(image: np.ndarray) -> np.ndarray:
    """Calculate Extended Depth of Field for the given input image Zstack.

    Args:
        image: Input image array of shape (Z, X, Y)

    Returns:
        Maximum Intensity Projected Image.
    """
    return np.max(image, axis=0)



from ._datasets import (
    autophagosome_h5sc,
    dataset_1,
    dataset_1_omezarr,
    dataset_2,
    dataset_3,
    dataset_4,
    dataset_5,
    dataset_6,
    dataset_7,
    dataset_parsing_example,
    dataset_parsing_example_flatfield_corrected,
    dataset_stitching_example,
    get_config_file,
)



import os
from typing import Any, Literal, TypeAlias

import numpy as np
from spatialdata import SpatialData, read_zarr
from spatialdata.models import get_model
from xarray import DataArray, DataTree

ObjectType: TypeAlias = Literal["images", "labels", "points", "tables", "shapes"]


def _get_image(elem: DataArray | DataTree) -> DataArray:
    """Get the image from the element.

    Args:
        elem: Element to get the image from

    Returns:
        Image of the element or None if the element is not an image
    """
    if isinstance(elem, DataArray):
        image = elem
    elif isinstance(elem, DataTree):
        image = elem.scale0.image
    else:
        image = None
    return image


def _get_shape(elem: DataArray | DataTree) -> tuple[int, int, int]:
    """Get the shape of the element.

    Args:
        elem: Element to get the shape of

    Returns:
        Tuple of the shape of the element with c, x, y dimensions. If the element is 2D, the first dimension is set to np.nan.
    """
    if isinstance(elem, DataArray):
        shape = elem.shape
    elif isinstance(elem, DataTree):
        shape = elem.scale0.image.shape
    else:
        raise ValueError(f"Element type {type(elem)} not supported.")

    if len(shape) == 2:
        shape = (np.nan, shape[0], shape[1])
    elif len(shape) == 3:
        shape = (shape[0], shape[1], shape[2])
    return shape


def _make_key_lookup(sdata: SpatialData) -> dict:
    """Make a lookup dictionary for the keys in the SpatialData object.

    Args:
        sdata: SpatialData object

    Returns:
        Dictionary of the keys in the SpatialData object
    """
    dict_lookup: dict[str, list[str]] = {}
    for elem in sdata.elements_paths_in_memory():
        key, name = elem.split("/")
        if key not in dict_lookup:
            dict_lookup[key] = []
        dict_lookup[key].append(name)
    return dict_lookup


def _force_delete_object(sdata: SpatialData, name: str) -> None:
    """Force delete an object from the SpatialData object and directory.

    Args:
        sdata: SpatialData object
        name: Name of object to delete
        type: Type of object ("images", "labels", "points", "tables")

    Returns:
        None the SpatialData object is updated on file
    """
    in_memory_only, _ = sdata._symmetric_difference_with_zarr_store()
    in_memory_only = [x.split("/")[-1] for x in in_memory_only]

    if name in sdata:
        del sdata[name]

    if name not in in_memory_only:
        sdata.delete_element_from_disk(name)


def add_element_sdata(sdata: SpatialData, element: Any, element_name: str, overwrite: bool = True):
    """Add an element to the SpatialData object.

    Args:
        sdata: SpatialData object
        element: Element to add
        element_name: Name of the element to be added
        overwrite: Whether to overwrite the element if it already exists

    Returns:
        None: the SpatialData object is updated on file
    """

    if element_name in sdata:
        if not overwrite:
            raise ValueError(
                f"Object with name '{element_name}' already exists in SpatialData." f"Set overwrite=True to replace it."
            )

        _force_delete_object(sdata, element_name)

    # the element needs to validate with exactly one of the models
    get_model(element)

    # Add the element to the SpatialData object
    sdata[element_name] = element
    sdata.write_element(element_name)


def rename_image_element(sdata: SpatialData, image_element: str, new_element_name: str) -> SpatialData:
    """Rename an image element in the sdata object.

    Args:
        sdata: SpatialData object
        image_element: Image element to rename
        new_element_name: New name for the image element

    Returns:
        sdata: Updated SpatialData object with the renamed image element.
    """
    assert image_element in sdata, f"Image element '{image_element}' not found in SpatialData."

    path_sdata = sdata.path
    path_elem = path_sdata / "images" / image_element
    new_path_elem = sdata.path / "images" / new_element_name

    # Check if the new name already exists
    if new_path_elem.exists():
        raise ValueError(f"Image element with name '{new_element_name}' already exists.")

    short_path_elem = f"images/{image_element}"
    short_path_new_elem = f"images/{new_element_name}"

    assert (
        str(short_path_elem) in sdata.elements_paths_on_disk()
    ), f"Element {image_element} needs to be on disk to rename it."

    # rename metadata
    zattrs_path = sdata.path / "images" / image_element / ".zattrs"
    with open(zattrs_path) as file:
        content = file.read()
        content = content.replace(f"/{short_path_elem}", f"/{short_path_new_elem}")

    with open(zattrs_path, "w") as file:
        file.write(content)

    # rename image files
    os.rename(path_elem, new_path_elem)

    # read and return update spatialdata object
    return read_zarr(path_sdata)



import itertools

import numpy as np
import pandas as pd
import torch
import wandb

# Check if wandb is installed as optional dependency
try:
    import wandb
except ImportError as err:
    raise ImportError(
        "Wandb is not installed. Please install it via `pip install wandb` "
        "if you want to utilize the callback metrics."
    ) from err

# Check if Pytorch Lightning is installed as optional dependency
try:
    from pytorch_lightning import LightningModule
    from pytorch_lightning.callbacks import Callback
    from pytorch_lightning.trainer import Trainer

except ImportError as err:
    raise ImportError(
        "Pytorch Lightning is not installed. Please install it via `pip install pytorch_lightning` "
        "if you want to utilize the callback metrics."
    ) from err

from sklearn.metrics import (
    auc,
    confusion_matrix,
    f1_score,
    precision_recall_curve,
    precision_score,
    roc_curve,
)


class BatchAccumulatedMetricsCallback(Callback):
    """
    Callback to calculate metrics on the accumulated predictions of a batch.

    The following metrics are calculated:
    - F1-score
    - AUC precision-recall
    - AUC ROC curve
    - Precision-recall curve
    - ROC curve

    The precision-recall curve and ROC curve are only calculated after a complete epoch to reduce the amount of data that needs to be saved.
    This metric relies on wandb for logging the data.
    """

    def __init__(self, downsampling_factor: int = 4, n_epochs_for_big_calcs: int = 1) -> None:
        """
        Args:
            downsampling_factor: Factor to downsample the roc and precision-recall curve data to reduce the amount of data that needs to be saved. Defaults to 4.
            n_epochs_for_big_calcs: Number of epochs after which to calculate the roc and precision-recall curve data. All other data is calculated at each validation step. Defaults to 1.

        """
        self.iteration = 1
        self.n_epochs_big_calcs = n_epochs_for_big_calcs
        self.downsampling_factor = downsampling_factor

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.train_actual_labels = torch.tensor([])
        self.train_probabilities = torch.tensor([])

    def on_train_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: dict, batch, batch_idx, dataloader_idx=0
    ) -> None:
        self.train_actual_labels = torch.cat((self.train_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.train_probabilities = torch.cat((self.train_probabilities, outputs["probabilities"].detach().cpu()))

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # calculate f1-score
        probabilities = self.train_probabilities.numpy()
        probs_1d = probabilities[:, 1]  # only get the predictions for the true-class
        preds_1d = (probs_1d >= 0.5).astype(int)
        labels = self.train_actual_labels.numpy()

        f1 = f1_score(y_true=labels, y_pred=preds_1d)
        self.log("f1_score/train_accumulated", f1, sync_dist=True)

    def on_validation_epoch_start(self, trainer, pl_module):
        self.val_actual_labels = torch.tensor([])
        self.val_probabilities = torch.tensor([])

    def on_validation_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: dict, batch, batch_idx, dataloader_idx=0
    ) -> None:
        self.val_actual_labels = torch.cat((self.val_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.val_probabilities = torch.cat((self.val_probabilities, outputs["probabilities"].detach().cpu()))

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # calculate f1-score
        probabilities = self.val_probabilities.numpy()
        probs_1d = probabilities[:, 1]  # only get the predictions for the true-class
        preds_1d = (probs_1d >= 0.5).astype(int)
        labels = self.val_actual_labels.numpy()

        f1 = f1_score(y_true=labels, y_pred=preds_1d)
        self.log("f1_score/val_accumulated", f1, sync_dist=True)

        # calculate precision-recall curve
        precision, recall, thresholds = precision_recall_curve(labels, probs_1d, pos_label=1)
        auc_precision_recall = auc(recall, precision)

        self.log("auc_precision_recall/val_accumulated", auc_precision_recall, sync_dist=True)

        # calculate ROC
        fpr, tpr, thresholds = roc_curve(labels, probs_1d, pos_label=1)
        auc_roc = auc(fpr, tpr)

        self.log("auc_roc/val_accumulated", auc_roc, sync_dist=True)

        # we only want to log the roc curve and precision recall curve after a complete epoch (not for each validation step)
        # other metrics should be logged at all val checks though
        if self.iteration % (1 / trainer.val_check_interval) * self.n_epochs_big_calcs == 0:
            # plot roc curve
            data = pd.DataFrame([[x, y] for (x, y) in zip(tpr, fpr, strict=True)])
            data = data.iloc[
                :: self.downsampling_factor, :
            ]  # subsample to every second entry to make amount of data that needs to be saved smaller
            data.columns = ["tpr", "fpr"]
            table = wandb.Table(data=data, columns=["tpr", "fpr"])
            trainer.logger.experiment.log({"roc_curve/val_accumulated": table, "t_epoch": trainer.current_epoch})

            # plot precision-recall curve
            data = [[x, y] for (x, y) in zip(precision, recall, strict=True)]
            data = pd.DataFrame(data)
            data = data.iloc[
                :: self.downsampling_factor, :
            ]  # subsample to every second entry to make amount of data that needs to be saved smaller
            data.columns = ["precision", "recall"]
            table = wandb.Table(data=data, columns=["precision", "recall"])
            trainer.logger.experiment.log(
                {"precision_recall_curve/val_accumulated": table, "t_epoch": trainer.current_epoch}
            )

        self.iteration = self.iteration + 1

    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.test_actual_labels = torch.tensor([])
        self.test_probabilities = torch.tensor([])

    def on_test_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: dict, batch, batch_idx, dataloader_idx: int = 0
    ) -> None:
        self.test_actual_labels = torch.cat((self.test_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.test_probabilities = torch.cat((self.test_probabilities, outputs["probabilities"].detach().cpu()))

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        # calculate f1-score
        probabilities = self.test_probabilities.numpy()
        probs_1d = probabilities[:, 1]  # only get the predictions for the true-class
        preds_1d = (probs_1d >= 0.5).astype(int)
        labels = self.test_actual_labels.numpy()

        f1 = f1_score(y_true=labels, y_pred=preds_1d)
        self.log("f1_score/test_accumulated", f1, sync_dist=True)

        # calculate precision-recall curve
        precision, recall, thresholds = precision_recall_curve(labels, probs_1d, pos_label=1)
        auc_precision_recall = auc(recall, precision)

        self.log("auc_precision_recall/test_accumulated", auc_precision_recall, sync_dist=True)

        # calculate ROC
        fpr, tpr, thresholds = roc_curve(labels, probs_1d, pos_label=1)
        auc_roc = auc(fpr, tpr)

        self.log("auc_roc/test_accumulated", auc_roc, sync_dist=True)

        # plot roc curve
        data = [[x, y] for (x, y) in zip(tpr, fpr, strict=True)]
        table = wandb.Table(data=data, columns=["tpr", "fpr"])
        trainer.logger.experiment.log({"roc_curve/test_accumulated": table, "t_epoch": trainer.current_epoch})

        # plot precision-recall curve
        data = [[x, y] for (x, y) in zip(precision, recall, strict=True)]
        table = wandb.Table(data=data, columns=["precision", "recall"])
        trainer.logger.experiment.log(
            {"precision_recall_curve/test_accumulated": table, "t_epoch": trainer.current_epoch}
        )


class MulticlassBatchAccumulatedMetricsCallback(Callback):
    """
    Multiclass variant of the callback to calculate metrics on the accumulated predictions of a batch.

    The following metrics are calculated:
    - Precision per class
    - Confusion matrix

    The confusion matrix is only calculated after a complete epoch to reduce the amount of data that needs to be saved.
    """

    def __init__(self, downsampling_factor: int = 4, n_epochs_for_big_calcs: int = 1) -> None:
        """
        Args:
            downsampling_factor (int, optional): Factor to downsample the roc and precision-recall curve data to reduce the amount of data that needs to be saved. Defaults to 4.
            n_epochs_for_big_calcs (int, optional): Number of epochs after which to calculate the roc and precision-recall curve data. All other data is calculated at each validation step. Defaults to 1.
        """
        self.iteration = 1
        self.n_epochs_big_calcs = n_epochs_for_big_calcs
        self.downsampling_factor = downsampling_factor
        pass

    def _calculate_confusion_table(
        self, y_true: torch.Tensor, y_pred: torch.Tensor, labels: torch.Tensor | None = None, normalize: bool = False
    ) -> wandb.Table:
        """Calculate a confusion matrix using the given true and predicted labels.

        Args:
            y_true: True labels.
            y_pred: Predicted labels.
            labels: Labels to use for the confusion matrix. Defaults to None.
            normalize: Whether to normalize the confusion matrix. Defaults to False.

        Returns:
            Table containing the confusion matrix data.
        """
        y_true = np.asarray(y_true)
        y_pred = np.asarray(y_pred)
        cm = confusion_matrix(y_true, y_pred)

        if labels is None:
            classes = np.unique((y_true, y_pred))
        else:
            classes = np.asarray(labels)

        if normalize:
            cm = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis]
            cm = np.around(cm, decimals=2)
            cm[np.isnan(cm)] = 0.0

        np.isin(classes, y_true)
        true_classes = classes

        np.isin(classes, y_pred)
        pred_classes = classes

        data = []
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            if labels is not None and (isinstance(pred_classes[i], int) or isinstance(pred_classes[0], np.integer)):
                pred_dict = labels[pred_classes[i]]
                true_dict = labels[true_classes[j]]
            else:
                pred_dict = pred_classes[j]
                true_dict = true_classes[i]

            data.append([pred_dict, true_dict, cm[i, j]])

        return wandb.Table(columns=["Predicted_Label", "Actual_Label", "Count"], data=data)

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.train_actual_labels = torch.tensor([])
        self.train_probabilities = torch.tensor([])

    def on_train_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: dict,
        batch: torch.Tensor,
        batch_idx: int,
        dataloader_idx: int = 0,
    ):
        self.train_actual_labels = torch.cat((self.train_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.train_probabilities = torch.cat((self.train_probabilities, outputs["probabilities"].detach().cpu()))

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        probabilities = self.train_probabilities.numpy()
        predictions = np.argmax(probabilities, axis=1).astype(int)

        labels = self.train_actual_labels.numpy()
        np.sort(np.unique(labels))

        precision_per_class = precision_score(labels, predictions, average=None)
        precision_dict = {f"precision_class_{i}/train": p for i, p in enumerate(precision_per_class)}
        trainer.logger.experiment.log(precision_dict)

    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.val_actual_labels = torch.tensor([])
        self.val_probabilities = torch.tensor([])

    def on_validation_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: dict,
        batch: torch.Tensor,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        self.val_actual_labels = torch.cat((self.val_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.val_probabilities = torch.cat((self.val_probabilities, outputs["probabilities"].detach().cpu()))

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        probabilities = self.val_probabilities.numpy()
        predictions = np.argmax(probabilities, axis=1).astype(int)

        labels = self.val_actual_labels.numpy()
        label_overview = np.sort(np.unique(labels))

        precision_per_class = precision_score(labels, predictions, average=None)
        precision_dict = {f"precision_class_{i}/val": p for i, p in enumerate(precision_per_class)}
        trainer.logger.experiment.log(precision_dict)

        if self.iteration % (1 / trainer.val_check_interval) * self.n_epochs_big_calcs == 0:
            for label in label_overview:
                filtered = probabilities[labels == label]
                means = np.mean(filtered, axis=0)

                data = [[label, val] for (label, val) in zip(label_overview, means, strict=False)]
                table = wandb.Table(data=data, columns=["label", "mean prediction"])
                trainer.logger.experiment.log(
                    {f"prediction distribution true class {label}/val": table, "t_epoch": trainer.current_epoch}
                )

            table = self._calculate_confusion_table(y_true=labels, y_pred=predictions)
            trainer.logger.experiment.log({"confusion matrix/val": table, "t_epoch": trainer.current_epoch})

        self.iteration = self.iteration + 1

    def on_test_epoch_start(self, trainer: Trainer, pl_module: LightningModule) -> None:
        self.test_actual_labels = torch.tensor([])
        self.test_probabilities = torch.tensor([])

    def on_test_batch_end(
        self, trainer: Trainer, pl_module: LightningModule, outputs: dict, batch, batch_idx, dataloader_idx: int = 0
    ) -> None:
        self.test_actual_labels = torch.cat((self.test_actual_labels, outputs["actual_labels"].detach().cpu()))
        self.test_probabilities = torch.cat((self.test_probabilities, outputs["probabilities"].detach().cpu()))

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        probabilities = self.test_probabilities.numpy()
        predictions = np.argmax(probabilities, axis=1).astype(int)

        labels = self.test_actual_labels.numpy()
        label_overview = np.sort(np.unique(labels))

        precision_per_class = precision_score(labels, predictions, average=None)
        precision_dict = {
            f"precision_class_{i}/val": p for i, p in zip(label_overview, precision_per_class, strict=False)
        }
        trainer.logger.experiment.log(precision_dict)

        if self.iteration % (1 / trainer.val_check_interval) * self.n_epochs_big_calcs == 0:
            for label in label_overview:
                filtered = probabilities[labels == label]
                means = np.mean(filtered, axis=0)

                data = [[label, val] for (label, val) in zip(label_overview, means, strict=False)]
                table = wandb.Table(data=data, columns=["label", "mean prediction"])
                trainer.logger.experiment.log(
                    {f"prediction distribution true class {label}/val": table, "t_epoch": trainer.current_epoch}
                )

            table = self._calculate_confusion_table(y_true=labels, y_pred=predictions)
            trainer.logger.experiment.log({"confusion matrix/test": table, "t_epoch": trainer.current_epoch})
